#line 120 "/home/dodo/tmp/BB/bff-1.1/src/main.u"
#include <stdio.h>
#include <time.h>
#include <stdlib.h>
#include <errno.h>
#include <string.h>
#include "parse.h"

#line 9 "main.c"

static void usage(char *prg);
static defile_t* parse_inputs(int argc,char **argv);
static void merge_files(defile_t *mf, defile_t *af);
static void merge_raws(raw_t *M, raw_t *A);
static void gen_typedef(FILE *out, defile_t *fil, odef_t *od);
static void gen_struct(FILE *out,defile_t *fil,odef_t *od);
static int gen_basic_type(FILE *out, defile_t *fil, type_t *T);
static void gen_decl(FILE *out, defile_t *fil, type_t *T, char *name,int end);
static void gen_all_struct(FILE *out,defile_t *fil);
static void gen_elementary_reader
   (FILE *out, defile_t *fil, type_t *T, char *ptr);
static void gen_array_loop(FILE *out, type_t *T);
static void gen_array_reference(FILE *out, char *d, type_t *T);
static void gen_field_reader
     (FILE *out, defile_t *fil, odef_t *od, field_t *f);
static int imax2(int a,int b) ;
static void gen_array_index_declarations(FILE *out, odef_t *od);
static void gen_decode_function
     (FILE *out, defile_t *fil, odef_t *od, int body);
static void gen_read_function
     (FILE *out, defile_t *fil, odef_t *od, int body);
static void gen_all_rw_function(FILE *hout, FILE *cout, defile_t *fil);
static void print_gen_time(FILE *out, struct tm *gt);
static void gen_unused_array(FILE *out, char *pfx);
static void gen_all(defile_t *df);
static int print_raw_lines(FILE *out, raw_t *raws);
static void gen_src_top(FILE *out, defile_t *df,struct tm *gti);
static void gen_src_bot(FILE *out, defile_t *df);
static void gen_hdr_top(FILE *out, defile_t *df,struct tm *gti);
static void gen_hdr_bot(FILE *out, defile_t *df);
static void gen_library(FILE *out, defile_t *df);
static void gen_write_function
     (FILE *out, defile_t *fil, odef_t *od, int body);
static void enumerate_objects(defile_t *fil);
static void gen_encode_function
    (FILE *out, defile_t *fil, odef_t *od,int body);
static void gen_field_writer
    (FILE *out, defile_t *fil, odef_t *od, field_t *f);
static void gen_elementary_writer
  (FILE *out, defile_t *fil, type_t *T, char *ref);
static void gen_magic(FILE *out, defile_t *fil);
static int proto_or_body(FILE *out, int body);
static int end_body(FILE *out);
static void gen_util_functions(FILE *out, defile_t *fil,int body);
void dbg_print_file(defile_t *df);
static void dbg_print_objects(odef_t *all);
static void dbg_print_type(type_t *t);
#line 131 "/home/dodo/tmp/BB/bff-1.1/src/main.u"

#line 10 "/home/dodo/tmp/BB/bff-1.1/src/main.u"

static void usage(char *prg)
{
  fprintf(stderr,"usage:\n");
  fprintf(stderr,"%s [-v] [-p prefix] [-s source] [-h header] input [ input ... ]\n",
   prg);
  exit(1);
}

static defile_t* parse_inputs(int argc,char **argv)
{
  defile_t *mf,*af,**t;
  int i;
  char *c_pfx, *c_src, *c_hdr;
  c_pfx= c_src= c_hdr= NULL;
  mf= NULL;
  for(i=1;i<argc;)
  {
    if (!strcmp(argv[i], "-v"))
    {
      printf("BFF version 1.1 - 3nov2017\n");
      exit(0);
    }
    if (!strcmp(argv[i],"-s"))
    {
      if (i==argc-1) usage(argv[0]);
      c_src= argv[i+1];
      i+= 2;
      continue;
    }
    if (!strcmp(argv[i],"-h"))
    {
      if (i==argc-1) usage(argv[0]);
      c_hdr= argv[i+1];
      i+= 2;
      continue;
    }
    if (!strcmp(argv[i],"-p"))
    {
      if (i==argc-1) usage(argv[0]);
      c_pfx= argv[i+1];
      i+= 2;
      continue;
    }
    if (mf) t= &af; else t= &mf;
    *t= parse_file(argv[i]); 

    if (t==&mf && !mf->has_magic)
      exit(fprintf(stderr,"%s: no magic was given\n",argv[i]));
    if (t==&af && af->has_magic)
      exit(fprintf(stderr,"%s: magic can be specified only in main file\n",
              argv[i]));
    i++;
    if (t==&af) { merge_files(mf, af); af= NULL; }
  }
  if (!mf) usage(argv[0]);
  if (c_src) mf->sname= c_src;
  if (c_hdr) mf->hname= c_hdr;
  if (c_pfx) mf->pfx= c_pfx;

  resolve_unknown(mf);
  return mf;
}

static void merge_files(defile_t *mf, defile_t *af)
{
  odef_t *odM,*odA;
  raw_t *f,*l;

  for(odM=mf->odefs.next;odM!=&mf->odefs;odM=odM->next)
    odM->raws.next= odM->raws.prev= &odM->raws;

  for(odA=af->odefs.next;odA!=&af->odefs;odA=odA->next)
  {
    for(odM=mf->odefs.next;odM!=&mf->odefs;odM=odM->next)
      if (!strcmp(odM->name,odA->name)) break;
    if (odM==&mf->odefs)
      exit(fprintf(stderr,"%s: new definition in auxilliary file "
                          "is not allowed\n", odA->src));
    if (odA->fields.next!=&odA->fields)
      exit(fprintf(stderr,"%s: field definitions are not allowed in "
                          "auxilliary files\n",odA->fields.next->src));

    f= odA->raws.next;
    l= odA->raws.prev;
    odM->raws.next= f;
    odM->raws.prev= l;
    f->prev= l->next= &odM->raws;
  }

  if (af->pfx) mf->pfx= af->pfx;
  if (af->sname) mf->sname= af->sname;
  if (af->hname) mf->hname= af->hname;

  merge_raws(&mf->src_top, &af->src_top);
  merge_raws(&mf->src_bot, &af->src_bot);
  merge_raws(&mf->hdr_top, &af->hdr_top);
  merge_raws(&mf->hdr_bot, &af->hdr_bot);
}

static void merge_raws(raw_t *M, raw_t *A)
{
  if (A->next==A) return ;
  M->next= A->next;
  M->prev= A->prev;
  M->next->prev= M;
  M->prev->next= M;
}

#line 8 "/home/dodo/tmp/BB/bff-1.1/src/gen.u"

static void gen_typedef(FILE *out, defile_t *fil, odef_t *od)
{
  fprintf(out,"typedef struct %s%s %s%s_t;\n", 
      fil->pfx, od->name, fil->pfx, od->name);
}

static void gen_struct(FILE *out,defile_t *fil,odef_t *od)
{
  field_t *f;
  raw_t *r;
  fprintf(out,"struct %s%s { \n", fil->pfx, od->name);
  for(f=od->fields.next;f!=&od->fields;f=f->next)
     gen_decl(out, fil, f->type, f->name, 1);
  for(r=od->raws.next;r!=&od->raws;r=r->next)
     fprintf(out,"  %s\n", r->text);
  if (od->top_level) fprintf(out,"  uint64_t _objid;\n");
  fprintf(out, "};\n");
}

static int gen_basic_type(FILE *out, defile_t *fil, type_t *T)
{
  odef_t *O;
  switch(T->kind)
  {
    case TYPE_I8: fprintf(out,"  int8_t "); break;
    case TYPE_U8: fprintf(out,"  uint8_t "); break;
    case TYPE_I16: fprintf(out,"  int16_t "); break;
    case TYPE_U16: fprintf(out,"  uint16_t "); break;
    case TYPE_I32: fprintf(out,"  int32_t "); break;
    case TYPE_U32: fprintf(out,"  uint32_t "); break;
    case TYPE_I64: fprintf(out,"  int64_t "); break;
    case TYPE_U64: fprintf(out,"  uint64_t "); break;
    case TYPE_F32: fprintf(out,"  float "); break;
    case TYPE_F64: fprintf(out,"  double "); break;
    case TYPE_S8: fprintf(out,"  uint8_t *"); break;
    case TYPE_S16: fprintf(out,"  uint16_t *"); break;
    case TYPE_S32: fprintf(out,"  uint32_t *"); break;
    case TYPE_SUB:
      O= T->obdef;
      fprintf(out,"  %s%s_t ", fil->pfx, O->name);
      break;
    default: return 0;
  }
  return 1;
}

static void gen_decl(FILE *out, defile_t *fil, type_t *T, char *name,int end)
{
  int i;
  char *K;
  odef_t *O;

  switch(T->kind)
  {
   case TYPE_ARRAY:
    gen_decl(out, fil, T->ref, name,0);
    for(i=0;i<T->n_dimensions;i++)
      fprintf(out,"[%d]", T->dimensions[i]);
    if (end) fprintf(out, ";\n");
    return ;
   case TYPE_DARRAY:
    fprintf(out,"  uint32_t N%s, S%s;\n", name, name);
    goto declare_pointer;
   case TYPE_POINTER:
    fprintf(out,"  uint64_t ID%s;\n", name);
    goto declare_pointer;
   default:
    if (gen_basic_type(out,fil,T)) 
    {
      fprintf(out," %s", name);
      if (end) fprintf(out,";\n");
    }
  }
  return ;
declare_pointer:
  K= malloc(strlen(name)+2);
  sprintf(K, "*%s",name);
  gen_decl(out, fil,T->ref, K,end);
  free(K);
}

static void gen_all_struct(FILE *out,defile_t *fil)
{
  odef_t *od;
  fprintf(out,"#ifndef _BFF_TYPE_DEFINED\n");
  fprintf(out,"#define _BFF_TYPE_DEFINED\n");
  fprintf(out,"typedef struct _bff bff_t;\n");
  fprintf(out,"#endif\n");
  for(od=fil->odefs.next;od!=&fil->odefs;od=od->next)
    gen_typedef(out, fil, od);

  for(od=fil->odefs.next;od!=&fil->odefs;od=od->next)
    gen_struct(out, fil, od);
}

static void gen_elementary_reader
   (FILE *out, defile_t *fil, type_t *T, char *ptr)
{
  char buf[200];
  if (type_to_str(T, buf)) 
  {
    odef_t *O;
    if (T->kind!=TYPE_SUB) return ;
    O= T->obdef;
    fprintf(out,"  %sdecode_%s(bf,%s);\n", fil->pfx, O->name, ptr);
    return ;
  }
  fprintf(out,"  bff_bread_%s(bf,%s);\n", buf, ptr);
}

static void gen_array_loop(FILE *out, type_t *T)
{
  int i;
  for(i=0;i<T->n_dimensions;i++)
    fprintf(out, "for(i%d=0;i%d<%d;i%d++)\n", i, i, T->dimensions[i], i);
}

static void gen_array_reference(FILE *out, char *d, type_t *T)
{
  int i;
  d+= strlen(d);
    for(i=0;i<T->n_dimensions;i++)
      { sprintf(d,"[i%d]", i); d+= strlen(d); }
}

static void gen_field_reader
     (FILE *out, defile_t *fil, odef_t *od, field_t *f)
{
  type_t *T;
  int i;
  char buf[2000];
  char *d;
  size_t K;
  T= f->type;
  switch(T->kind)
  {
   
  case TYPE_ARRAY:
    gen_array_loop(out, T);
    d= buf;
    sprintf(buf,"&OBJ->%s", f->name); 
    gen_array_reference(out, buf, T);
    gen_elementary_reader(out, fil, T->ref, buf);
    return ;
  case TYPE_POINTER:
    fprintf(out,"  bff_bread_u64(bf, &OBJ->ID%s);\n", f->name); 
    return ;
  case TYPE_DARRAY:
    fprintf(out,"  bff_bread_u32(bf, &OBJ->N%s);\n", f->name);
    fprintf(out,"  OBJ->S%s= OBJ->N%s;\n", f->name, f->name);
    fprintf(out,"  OBJ->%s= bff_ralloc(bf,sizeof(OBJ->%s[0])*OBJ->N%s);\n",
                       f->name, f->name, f->name);
    fprintf(out,"  for(i0=0;i0<OBJ->N%s;i0++)\n", f->name);
    sprintf(buf, "&OBJ->%s[i0]", f->name);
    gen_elementary_reader(out, fil, T->ref, buf);
    return ;
  }
  sprintf(buf, "&OBJ->%s", f->name);
  gen_elementary_reader(out, fil, T, buf);
}

static int imax2(int a,int b) { return a>b?a:b; }

static void gen_array_index_declarations(FILE *out, odef_t *od)
{
  int maxi,i;
  field_t *f;

  maxi= 0;
  for(f=od->fields.next;f!=&od->fields;f=f->next)
   switch(f->type->kind)
   {
   case TYPE_DARRAY: maxi=  imax2(maxi, 1); break;
   case TYPE_ARRAY: maxi= imax2(maxi,f->type->n_dimensions); break;
   }

  if (maxi)
  {
    fprintf(out,"  int");
    for(i=0;i<maxi;i++) fprintf(out,"%ci%d", i? ',':' ', i);
    fprintf(out,";\n");
  }
}

static void gen_decode_function
     (FILE *out, defile_t *fil, odef_t *od, int body)
{
  field_t *f;
  fprintf(out,"static void %sdecode_%s(bff_t *bf, %s%s_t *OBJ)",
               fil->pfx, od->name, fil->pfx, od->name);
  if (!body) { fprintf(out,";\n"); return ; }
  fprintf(out,"\n{\n");

  gen_array_index_declarations(out, od);

  for(f=od->fields.next;f!=&od->fields;f=f->next)
    gen_field_reader(out, fil, od, f);
  fprintf(out,"}\n");
}

static void gen_read_function
     (FILE *out, defile_t *fil, odef_t *od, int body)
{
  field_t *f;
  fprintf(out,"int %sread_%s(bff_t *bf, uint64_t objid, %s%s_t **R)",
               fil->pfx, od->name, fil->pfx, od->name);
  if (!body) { fprintf(out,";\n"); return ; }
  fprintf(out,"\n{\n");
  fprintf(out,"  %s%s_t *OBJ;\n",fil->pfx,od->name);

  fprintf(out,"  if (bff_read_object(bf,objid)) return 1;\n");
  fprintf(out,"  if (bf->o_type!=0x%x) return 1;\n", od->typenumber);
  fprintf(out,"  OBJ= bff_ralloc0(bf, sizeof(*OBJ));\n");
  fprintf(out,"  OBJ->_objid= objid;\n");
  fprintf(out,"  if (setjmp(bf->ejump)) return 1;\n");
  fprintf(out,"  %sdecode_%s(bf, OBJ);\n", fil->pfx, od->name);

  fprintf(out,"  *R= OBJ;\n");
  fprintf(out,"  bff_reset_ralloc(bf);\n");
  fprintf(out,"  return 0;\n");
  fprintf(out,"}\n");
}

static void gen_all_rw_function(FILE *hout, FILE *cout, defile_t *fil)
{
  odef_t *od;
  int body;
  for(body=0;body<2;body++)
  for(od=fil->odefs.next;od!=&fil->odefs;od=od->next)
  {
   gen_decode_function(cout,fil,od,body);
   if (od->top_level)
     gen_read_function(cout,fil,od,body);
   gen_encode_function(cout,fil,od,body);
   if (od->top_level)
     gen_write_function(cout,fil,od,body);
  }
  for(od=fil->odefs.next;od!=&fil->odefs;od=od->next)
  {
   if (od->top_level)
     gen_read_function(hout,fil,od,0);
   if (od->top_level)
     gen_write_function(hout,fil,od,0);
  }

}


#line 261 "/home/dodo/tmp/BB/bff-1.1/src/gen.u"

static void print_gen_time(FILE *out, struct tm *gt)
{
  char buf[2000];
  strftime(buf, sizeof(buf), "%F %R", gt);
  fprintf(out, "/** Generated by BFF on %s **/\n", buf);
}

static void gen_unused_array(FILE *out, char *pfx)
{
  fprintf(out,"void *__%s_unused[]={\n", pfx);
  fprintf(out,"(void*) fbm_find, \n");
  fprintf(out,"(void*) postpone_buffer_free, \n");
  fprintf(out,"(void*) postpone_buffer_append, \n");
  fprintf(out,"(void*) postpone_buffer_write, \n");
  fprintf(out,"};\n");
}

static void gen_all(defile_t *df)
{
  FILE *h_out, *c_out;
  struct tm *gt;
  time_t tt;

  tt= time(NULL);
  gt= gmtime(&tt);

  h_out= fopen(df->hname,"w");
  if (!h_out) 
     exit(fprintf(stderr,"can't open %s: %s\n",
          df->hname, strerror(errno)));
  c_out= fopen(df->sname,"w");
  if (!c_out) 
     exit(fprintf(stderr,"can't open %s: %s\n",
          df->sname, strerror(errno)));
  gen_src_top(c_out, df,gt);
  gen_hdr_top(h_out, df,gt);

  gen_all_struct(c_out, df);
  gen_library(c_out, df);
  gen_magic(c_out, df);

  gen_all_struct(h_out, df);

  gen_all_rw_function(h_out, c_out, df);

  gen_util_functions(h_out, df, 0);
  gen_util_functions(c_out, df, 0);
  gen_util_functions(c_out, df, 1);

  gen_unused_array(c_out, df->pfx);

  gen_src_bot(c_out, df);
  gen_hdr_bot(h_out, df);

  fclose(h_out);
  fclose(c_out);
}

static int print_raw_lines(FILE *out, raw_t *raws)
{
  raw_t *r;
  if (raws->next==raws) return 0;
  for(r=raws->next;r!=raws;r=r->next)
    fprintf(out,"%s", r->text);
  return 1;
}

static void gen_src_top(FILE *out, defile_t *df,struct tm *gti)
{
  print_gen_time(out, gti);
  if (print_raw_lines(out, &df->src_top)) return ;
  fprintf(out,"%s",
#line 493 "main.c"
"#include <stdlib.h>\n"
"#include <stdio.h>\n"
"#include <stdint.h>\n"
"#include <stdarg.h>\n"
"#include <string.h>\n"
"#include <fcntl.h>\n"
"#include <unistd.h>\n"
"#include <errno.h>\n"
"#include <setjmp.h>\n"

#line 337 "/home/dodo/tmp/BB/bff-1.1/src/gen.u"
  ); 
}

static void gen_src_bot(FILE *out, defile_t *df)
{
  print_raw_lines(out, &df->src_bot);
}

static void gen_hdr_top(FILE *out, defile_t *df,struct tm *gti)
{
  print_gen_time(out, gti);
  print_raw_lines(out, &df->hdr_top);
}

static void gen_hdr_bot(FILE *out, defile_t *df)
{
  print_raw_lines(out, &df->hdr_bot);
}

static void gen_library(FILE *out, defile_t *df)
{
  fprintf(out,"%s",
#line 526 "main.c"
"\n"
"#define BLOCK_HEADER_SIZE     24\n"
"#define OBJECT_HEADER_SIZE     8\n"
"#define FREE_BLOCK_THRESHOLD  (2*(BLOCK_HEADER_SIZE+OBJECT_HEADER_SIZE))\n"
"#define FILE_HEADER_SIZE      32\n"
"\n"
"\n"
"typedef struct { uint64_t offset, size; int qindex; } free_block_t;\n"
"typedef struct { uint64_t ident, type, offset; } object_descriptor_t;\n"
"typedef struct { void *obj; int (*func)(void *, void *); } postponed_write_t;\n"
"\n"
"typedef struct fbm_tree fbm_t;\n"
"\n"
"\n"
"static fbm_t *fbm_new    ();\n"
"static free_block_t  *fbm_find   (fbm_t *T, uint64_t key);\n"
"static free_block_t  *fbm_insert (fbm_t *T, free_block_t *data);\n"
"static free_block_t  *fbm_remove (fbm_t *T, uint64_t key);\n"
"static free_block_t  *fbm_destroy(fbm_t *T);\n"
"\n"
"\n"
"\n"
"typedef struct fbm_node {\n"
"   struct fbm_node *left, *right,*parent;\n"
"   int height;\n"
"   free_block_t *data;\n"
"} fbm_node_t;\n"
"\n"
"struct fbm_tree { fbm_node_t *root; };\n"
"\n"
"\n"
"static int fbm_compare(uint64_t A, uint64_t B) { return (A<B?-1:(A>B?1:0)); }\n"
"static uint64_t fbm_upper(free_block_t *A)        { return (A->offset+A->size-1); }\n"
"static uint64_t fbm_lower(free_block_t *A)        { return A->offset; }\n"
"static int fbm_HEIGHT(fbm_node_t * A) { return A? A->height : 0; }\n"
"\n"
"static void fbm_calculate_height(fbm_node_t *N)\n"
"{\n"
"  int a,b;\n"
"  a= fbm_HEIGHT(N->left);\n"
"  b= fbm_HEIGHT(N->right);\n"
"  N->height= (a>b?a:b)+1;\n"
"}\n"
"\n"
"static void fbm_replace_child_in_parent \n"
"      (fbm_t* T, fbm_node_t* old, fbm_node_t* nw)\n"
"{\n"
"  if (old->parent)\n"
"    {\n"
"      if (old->parent->left == old)\n"
"\011old->parent->left = nw;\n"
"      else\n"
"\011old->parent->right = nw;\n"
"    }\n"
"  else\n"
"    {\n"
"      T->root = nw;\n"
"    }\n"
"  if (nw) nw->parent= old->parent;\n"
"}\n"
"\n"
"static void fbm_set_left_child(fbm_node_t *P, fbm_node_t *N)\n"
"{\n"
"  P->left= N; \n"
"  if (N) N->parent= P;\n"
"}\n"
"\n"
"static void fbm_set_right_child(fbm_node_t *P, fbm_node_t *N)\n"
"{\n"
"  P->right= N; \n"
"  if (N) N->parent= P;\n"
"}\n"
"\n"
"static fbm_node_t *fbm_find_node(fbm_t *T, uint64_t key)\n"
"{\n"
"   fbm_node_t *n;\n"
"   n= T->root;\n"
"   while (n)\n"
"   {\n"
"     int u;\n"
"     u= fbm_compare(key, fbm_lower(n->data));\n"
"     if (u==0) return n;\n"
"     if (u<0)  { n= n->left; continue; }\n"
"     u= fbm_compare(key, fbm_upper(n->data));\n"
"     if (u<=0) return n;\n"
"     n= n->right;\n"
"   }\n"
"   return NULL;\n"
"}\n"
"\n"
"static int fbm_right_heavy(fbm_node_t *A)\n"
"{\n"
"  if (!A) return 0;\n"
"  return (fbm_HEIGHT(A->right)-fbm_HEIGHT(A->left)) > 0;\n"
"}\n"
"\n"
"static int fbm_left_heavy(fbm_node_t *A)\n"
"{\n"
"  if (!A) return 0;\n"
"  return (fbm_HEIGHT(A->right)-fbm_HEIGHT(A->left)) < 0;\n"
"}\n"
"\n"
"static void fbm_rotate_left(fbm_t *T, fbm_node_t *A)\n"
"{\n"
"  fbm_node_t *B;\n"
"  B = A->right;\n"
"  fbm_set_right_child(A, B->left);\n"
"  fbm_replace_child_in_parent(T, A, B);\n"
"  fbm_set_left_child(B, A);\n"
"  fbm_calculate_height(A);\n"
"  fbm_calculate_height(B);\n"
"}\n"
"\n"
"static void fbm_rotate_left_twolevels(fbm_t *T, fbm_node_t *A)\n"
"{\n"
"  fbm_node_t *B, *C;\n"
"\n"
"  B = A->right; C = B->left;\n"
"  fbm_set_right_child(A, C->left);\n"
"  fbm_set_left_child(B, C->right);\n"
"  fbm_replace_child_in_parent(T, A, C);\n"
"  fbm_set_left_child(C, A);\n"
"  fbm_set_right_child(C, B);\n"
"  fbm_calculate_height(A);\n"
"  fbm_calculate_height(B);\n"
"  fbm_calculate_height(C);\n"
"}\n"
"\n"
"static void fbm_rotate_right(fbm_t *T, fbm_node_t *A)\n"
"{\n"
"  fbm_node_t *B;\n"
"  B = A->left;\n"
"  fbm_set_left_child(A, B->right);\n"
"  fbm_replace_child_in_parent(T, A, B);\n"
"  fbm_set_right_child(B, A);\n"
"  fbm_calculate_height(A);\n"
"  fbm_calculate_height(B);\n"
"}\n"
"\n"
"static void fbm_rotate_right_twolevels(fbm_t *T, fbm_node_t *A)\n"
"{\n"
"  fbm_node_t *B, *C;\n"
"  B = A->left; C = B->right;\n"
"  fbm_set_left_child(A, C->right);\n"
"  fbm_set_right_child(B, C->left);\n"
"  fbm_replace_child_in_parent(T, A, C);\n"
"  fbm_set_left_child(C, B);\n"
"  fbm_set_right_child(C, A);\n"
"  fbm_calculate_height(A);\n"
"  fbm_calculate_height(B);\n"
"  fbm_calculate_height(C);\n"
"}\n"
"\n"
"static void fbm_rebalance(fbm_t *T, fbm_node_t *N)\n"
"{\n"
"  int u;\n"
"  fbm_node_t *P;\n"
"\n"
"  fbm_calculate_height(N);\n"
"\n"
"  u = fbm_HEIGHT(N->left)-fbm_HEIGHT(N->right);\n"
"  P = N->parent;\n"
"  if (u < -1) {\n"
"      if (fbm_left_heavy(N->right)) fbm_rotate_left_twolevels(T, N);\n"
"                                else fbm_rotate_left(T, N);\n"
"  } else if (u > 1) {\n"
"      if (fbm_right_heavy(N->left)) fbm_rotate_right_twolevels(T, N);\n"
"                                else fbm_rotate_right(T, N);\n"
"  }\n"
"\n"
"  if (P) fbm_rebalance(T, P);\n"
"}\n"
"\n"
"static fbm_node_t *fbm_remove_max(fbm_t *T,fbm_node_t *N)\n"
"{\n"
"   while(N->right) N= N->right;\n"
"   fbm_replace_child_in_parent(T, N, N->left);\n"
"   if (N->parent) fbm_rebalance(T,N->parent);\n"
"   return N;\n"
"}\n"
"\n"
"static fbm_node_t *fbm_remove_min(fbm_t *T,fbm_node_t *N)\n"
"{\n"
"   while(N->left) N= N->left;\n"
"   fbm_replace_child_in_parent(T, N, N->right);\n"
"   if (N->parent) fbm_rebalance(T,N->parent);\n"
"   return N;\n"
"}\n"
"\n"
"static void *fbm_remove_node(fbm_t *T, fbm_node_t *N)\n"
"{\n"
"   fbm_node_t *Z;\n"
"   void *R;\n"
"   if (!N->left && !N->right) {\n"
"     fbm_replace_child_in_parent(T, N, NULL);\n"
"     if (N->parent) fbm_rebalance(T,N->parent);\n"
"   } else {\n"
"     if (fbm_left_heavy(N)) Z= fbm_remove_max(T,N->left);\n"
"                        else Z= fbm_remove_min(T,N->right);\n"
"     fbm_replace_child_in_parent(T, N, Z);\n"
"     fbm_set_left_child(Z, N->left);\n"
"     fbm_set_right_child(Z, N->right);\n"
"     fbm_rebalance(T,Z);\n"
"   }\n"
"   R= N->data;\n"
"   free(N);\n"
"   return R;\n"
"}\n"
"\n"
"\n"
"static fbm_t *fbm_new()\n"
"{\n"
"  fbm_t *T;\n"
"  T= malloc(sizeof(*T));\n"
"  T->root= NULL;\n"
"  return T;\n"
"}\n"
"\n"
"static free_block_t *fbm_find(fbm_t *T, uint64_t key)\n"
"{\n"
"   fbm_node_t *n;\n"
"   n= fbm_find_node(T, key);\n"
"   if (n) return n->data;\n"
"   return NULL;\n"
"}\n"
"\n"
"static free_block_t* fbm_insert(fbm_t *T, free_block_t *data)\n"
"{\n"
"  fbm_node_t *N,*P,*K,**PL;\n"
"  int R;\n"
"  uint64_t lower; \n"
"  uint64_t upper;\n"
"\n"
"  lower= fbm_lower(data);\n"
"  upper= fbm_upper(data);\n"
"  if (fbm_compare(lower,upper)>0) return data;\n"
"\n"
"  N = calloc(1, sizeof(*N));\n"
"  N->data= data;\n"
"  N->height= 1;\n"
"\n"
"  P = 0;\n"
"  PL= &T->root;\n"
"  K= T->root;\n"
"  while (K) {\n"
"    R= fbm_compare(upper, fbm_lower(K->data));\n"
"    if (R==0) { free(N); return K->data; }\n"
"    if (R<0)  { PL= &K->left; P= K; K= K->left; continue; }\n"
"    R= fbm_compare(lower, fbm_upper(K->data));\n"
"    if (R>0) { PL= &K->right; P= K; K= K->right; continue; }\n"
"    else     { free(N); return K->data; }\n"
"  }\n"
"  *PL= N;\n"
"  N->parent= P;\n"
"  fbm_rebalance(T, N);\n"
"  return NULL;\n"
"}\n"
"\n"
"static free_block_t *fbm_remove(fbm_t *T, uint64_t key)\n"
"{\n"
"   fbm_node_t *N;\n"
"   N= fbm_find_node(T,key);\n"
"   if (!N) return NULL;\n"
"   return fbm_remove_node(T,N);\n"
"}\n"
"\n"
"static free_block_t *fbm_destroy(fbm_t *T)\n"
"{\n"
"   fbm_node_t *N= T->root;\n"
"   if (!N) return NULL;\n"
"   while(N->left) N=N->left;\n"
"   return fbm_remove_node(T,N);\n"
"}\n"
"\n"
"\n"
"typedef struct postpone_buffer_buffer postpone_buffer_t;\n"
"\n"
"static postpone_buffer_t *postpone_buffer_new     (size_t blksize);\n"
"static void    postpone_buffer_free    (postpone_buffer_t *B, postponed_write_t **Rd, size_t *Rl);\n"
"static void    postpone_buffer_append  (postpone_buffer_t *dst, postpone_buffer_t *src_f);\n"
"static void    postpone_buffer_put     (postpone_buffer_t *dst, postponed_write_t d);\n"
"static void    postpone_buffer_write   (postpone_buffer_t *B, postponed_write_t *d, size_t l);\n"
"static size_t  postpone_buffer_read    (postpone_buffer_t *B, postponed_write_t *d, size_t l);\n"
"static void    postpone_buffer_collect (postpone_buffer_t *B, postponed_write_t **Rd, size_t *Rl);\n"
"\n"
"struct postpone_buffer_node { \n"
"  struct postpone_buffer_node *next;\n"
"  size_t len,size,start;\n"
"  postponed_write_t data[1];\n"
"};\n"
"\n"
"struct postpone_buffer_buffer {\n"
"  struct postpone_buffer_node *first, *last;\n"
"  size_t blksize;\n"
"};\n"
"\n"
"\n"
"\n"
"static struct postpone_buffer_node *postpone_buffer_get_free_node(postpone_buffer_t *B)\n"
"{\n"
"  struct postpone_buffer_node *P;\n"
"  \n"
"  P= B->last;\n"
"  if (P && (P->start+P->len)<P->size) \n"
"    return P;\n"
"\n"
"  P= malloc(sizeof(*P)+(B->blksize-1)*sizeof(postponed_write_t));\n"
"  P->len= 0;\n"
"  P->next= NULL;\n"
"  P->size= B->blksize;\n"
"  P->start= 0;\n"
"  if (!B->last) { B->last= B->first= P; }\n"
"           else { B->last->next= P; B->last= P; }\n"
"  return P;\n"
"}\n"
"\n"
"\n"
"static postpone_buffer_t *postpone_buffer_new(size_t blksize)\n"
"{\n"
"  postpone_buffer_t *B;\n"
"  B= malloc(sizeof(*B));\n"
"  B->first= B->last= NULL;\n"
"  B->blksize= blksize;\n"
"  return B;\n"
"}\n"
"\n"
"\n"
"static void postpone_buffer_free(postpone_buffer_t *B, postponed_write_t **Rd, size_t *Rl)\n"
"{\n"
"  struct postpone_buffer_node *P,*N;\n"
"  if (Rd || Rl) postpone_buffer_collect(B, Rd, Rl);\n"
"  for(P=B->first;P;P=N)\n"
"  {\n"
"    N= P->next;\n"
"    free(P);\n"
"  }\n"
"  free(B);\n"
"}\n"
"\n"
"static void postpone_buffer_collect(postpone_buffer_t *B, postponed_write_t **Rd, size_t *Rl)\n"
"{\n"
"  size_t S;\n"
"  struct postpone_buffer_node *P;\n"
"  postponed_write_t *R;\n"
"\n"
"  if (!Rd && !Rl) return ;\n"
"\n"
"  S= 0;\n"
"  for(P=B->first;P;P=P->next) S+= P->len;\n"
"\n"
"  if (S==0) { if (Rd) *Rd= NULL; if (Rl) *Rl= 0; return ; }\n"
"  if (!Rd) { *Rl= S; return ; }\n"
"\n"
"  R= malloc(S*sizeof(postponed_write_t));\n"
"  S= 0;\n"
"  for(P=B->first;P;P=P->next)\n"
"  {\n"
"    memcpy(R+S, P->data+P->start, P->len*sizeof(postponed_write_t));\n"
"    S+= P->len;\n"
"  }\n"
"\n"
"  *Rd= (void*) R;\n"
"  if (Rl) *Rl= S;\n"
"}\n"
"\n"
"static void postpone_buffer_write(postpone_buffer_t *B, postponed_write_t *d, size_t l)\n"
"{\n"
"  struct postpone_buffer_node *P;\n"
"  size_t cp;\n"
"\n"
"  while(l)\n"
"  {\n"
"    P= postpone_buffer_get_free_node(B);\n"
"    cp= P->size-P->len-P->start;\n"
"    if (cp>l) cp= l; \n"
"\n"
"    memcpy(P->data+P->len+P->start, d, cp*sizeof(postponed_write_t));\n"
"    P->len+= cp;\n"
"    l-= cp;\n"
"    d+= cp;\n"
"  }\n"
"}\n"
"\n"
"static void postpone_buffer_put(postpone_buffer_t *B, postponed_write_t d)\n"
"{\n"
"  struct postpone_buffer_node *P;\n"
"  P= postpone_buffer_get_free_node(B);\n"
"  P->data[P->start+P->len++]= d;\n"
"}\n"
"\n"
"static void postpone_buffer_append(postpone_buffer_t *dst, postpone_buffer_t *src_f)\n"
"{\n"
"  if (src_f->first!=NULL)\n"
"  {\n"
"    if (dst->first==NULL)\n"
"    {\n"
"      dst->first= src_f->first;\n"
"      dst->last= src_f->last;\n"
"    }\n"
"    else\n"
"    {\n"
"      dst->last->next= src_f->first;\n"
"      dst->last= src_f->last;\n"
"    }\n"
"  }\n"
"  free(src_f);\n"
"}\n"
"\n"
"static size_t postpone_buffer_read(postpone_buffer_t *B, postponed_write_t *d, size_t l)\n"
"{\n"
"  size_t total;\n"
"  struct postpone_buffer_node *P;\n"
"  size_t cp;\n"
"  total= 0;\n"
"\n"
"  while(l)\n"
"  {\n"
"    P= B->first;\n"
"    if (!P) break;  \n"
"    if (P->len==0)\n"
"    {\n"
"      B->first= P->next;\n"
"      if (!B->first) B->last= NULL;\n"
"      free(P);\n"
"      continue;\n"
"    } \n"
"    cp= P->len;\n"
"    if (cp>l) cp= l;\n"
"    memcpy(d, P->data+P->start, cp*sizeof(postponed_write_t));\n"
"    P->start+= cp;\n"
"    P->len-= cp;\n"
"    total+= cp;\n"
"    d+= cp;\n"
"  }\n"
"  return total;\n"
"}\n"
"\n"
"\n"
"typedef struct fbq_s fbq_t;\n"
"\n"
"\n"
"static fbq_t *fbq_new  (int initsize);\n"
"static void    fbq_free (fbq_t *q);\n"
"static void    fbq_put  (fbq_t *q, free_block_t* d);\n"
"static free_block_t*  fbq_get  (fbq_t *q);\n"
"static free_block_t*  fbq_peek (fbq_t *q);\n"
"\n"
"\n"
"static void fbq_remove(fbq_t *q, free_block_t* d);\n"
"\n"
"\n"
"\n"
"struct fbq_s\n"
"{\n"
"  size_t len,siz;\n"
"  free_block_t* *data;\n"
"};\n"
"\n"
"\n"
"static int fbq_max(fbq_t *q, int idx1, int idx2)\n"
"{\n"
"  int C;\n"
"  free_block_t* A;\n"
"  free_block_t* B;\n"
"  A= q->data[idx1];\n"
"  B= q->data[idx2];\n"
"  C= (A->size<B->size?-1:(A->size>B->size?1:0));\n"
"  if (C>0) return idx1;\n"
"  return idx2;\n"
"}\n"
"\n"
"static void fbq_exch(fbq_t *q,int a,int b)\n"
"{\n"
"  free_block_t* t;\n"
"  t= q->data[a];\n"
"  q->data[a]= q->data[b];\n"
"  q->data[b]= t;\n"
"  q->data[a]->qindex= a;\n"
"  q->data[b]->qindex= b;\n"
"}\n"
"\n"
"static int fbq_bubble_up(fbq_t *q,int idx)\n"
"{\n"
"  int P,G,S;\n"
"  int v;\n"
"  v= 0;\n"
"\n"
"again:\n"
"  if (idx==0) return v;\n"
"\n"
"  P= (idx-1)/2;\n"
"  S= 2*P+1;\n"
"  if (S==idx) S++;\n"
"  if (S>=q->len) G= idx;\n"
"            else G= fbq_max(q, S, idx);\n"
"\n"
"  if (fbq_max(q, P, G)==P) return v;\n"
"  fbq_exch(q,P,G);\n"
"  v++;\n"
"  idx= P;\n"
"  goto again;\n"
"}\n"
"\n"
"static int fbq_bubble_down(fbq_t *q, int idx)\n"
"{\n"
"  int L,R,G;\n"
"  int v;\n"
"  v= 0;\n"
"\n"
"again:\n"
"  L= 2*idx+1;\n"
"  if (L>=q->len) return v;\n"
"  R= L+1;\n"
"  if (R>=q->len) G= L;\n"
"            else G= fbq_max(q,L,R);\n"
"  if (fbq_max(q,G,idx)==idx) return v;\n"
"  fbq_exch(q,G,idx);\n"
"  idx= G;\n"
"  v++;\n"
"  goto again;\n"
"}\n"
"\n"
"\n"
"static fbq_t *fbq_new(int initsize)\n"
"{\n"
"  fbq_t *q;\n"
"  q= calloc(1,sizeof(*q));\n"
"  q->len= 0;\n"
"  if (initsize<2) initsize= 2;\n"
"  q->siz= initsize;\n"
"  q->data= malloc(sizeof(q->data[0])*q->siz);\n"
"  return q;\n"
"}\n"
"\n"
"static void fbq_free(fbq_t *q)\n"
"{\n"
"  free(q->data);\n"
"  free(q);\n"
"}\n"
"\n"
"static void fbq_put(fbq_t *q, free_block_t* d)\n"
"{\n"
"  if (q->len==q->siz)\n"
"  {\n"
"    q->siz+= q->siz/2;\n"
"    q->data= realloc(q->data,sizeof(q->data[0])*q->siz);\n"
"  }\n"
"  q->data[q->len]= d;\n"
"  d->qindex= q->len;\n"
"  q->len++;\n"
"  fbq_bubble_up(q, q->len-1);\n"
"}\n"
"\n"
"static free_block_t* fbq_get(fbq_t *q)\n"
"{\n"
"  free_block_t* d;\n"
"  if (q->len==0) return NULL;\n"
"  d= q->data[0];\n"
"  q->data[0]= q->data[q->len-1];\n"
"  q->data[0]->qindex= 0;\n"
"  q->len--;\n"
"  fbq_bubble_down(q, 0);\n"
"  return d;\n"
"}\n"
"\n"
"static free_block_t* fbq_peek(fbq_t *q)\n"
"{\n"
"  if (q->len==0) return NULL;\n"
"  return q->data[0];\n"
"}\n"
"\n"
"\n"
"\n"
"\n"
"\n"
"static void fbq_remove(fbq_t *q, free_block_t* d)\n"
"{\n"
"  int K;\n"
"  K= d->qindex;\n"
"  if (K<0 || K>=q->len) return ;\n"
"  q->len--;\n"
"  if (K==q->len) return ;\n"
"  q->data[K]= q->data[q->len];\n"
"  q->data[K]->qindex= K;\n"
"  (void) (fbq_bubble_down(q, K) || fbq_bubble_up(q, K));\n"
"}\n"
"\n"
"\n"
"\n"
"\n"
"typedef struct { void *root; } qumap_t;\n"
"\n"
"\n"
"static void     *qumap_find(qumap_t *map,uint64_t key);\n"
"static void     *qumap_insert(qumap_t* map, uint64_t key, void *data);\n"
"static void     *qumap_remove(qumap_t* map, uint64_t key);\n"
"static void     *qumap_destroy(qumap_t *map);\n"
"static qumap_t  *qumap_new();\n"
"static void      qumap_free(qumap_t *map);\n"
"\n"
"\n"
"#ifndef _QUMAP_KEY\n"
"#define _QUMAP_KEY(X)    ((X)->d.leaf.key)\n"
"#define _QUMAP_DATA(X)   ((X)->d.leaf.data)\n"
"#define _QUMAP_CHLD(X,Y) ((X)->d.inte.child[Y])\n"
"#endif\n"
"\n"
"\n"
"typedef struct\n"
"{\n"
"  union { struct { void* child[2];           } inte;  \n"
"          struct { void *data; uint64_t key; } leaf; } d;\n"
"  unsigned internal:1, bit:6;\n"
"} qumap_node_t;\n"
"\n"
"\n"
"static inline int qumap_dir(uint64_t key,qumap_node_t *node)\n"
"{\n"
"  return (key >> node->bit)&1;\n"
"}\n"
"\n"
"\n"
"static inline qumap_node_t* qumap_leaf_node(uint64_t key, void *data)\n"
"{\n"
"  qumap_node_t *N;\n"
"  N= malloc(sizeof(*N));\n"
"  N->internal= 0;\n"
"  _QUMAP_KEY(N)= key;\n"
"  _QUMAP_DATA(N)= data; \n"
"  return N;\n"
"}\n"
"\n"
"static inline qumap_node_t* qumap_internal_node(uint32_t bit)\n"
"{\n"
"  qumap_node_t *N;\n"
"  N= malloc(sizeof(*N));\n"
"  N->internal= 1;\n"
"  N->bit= bit;\n"
"  return N;\n"
"}\n"
"\n"
"static uint8_t qumap_crit_bit(uint64_t X, uint64_t Y)\n"
"{\n"
"  uint32_t i,j;\n"
"  static const uint64_t B[]=\n"
"  {\n"
"    0x5555555555555555ll,\n"
"    0x3333333333333333ll,\n"
"    0x0f0f0f0f0f0f0f0fll,\n"
"    0x00ff00ff00ff00ffll,\n"
"    0x0000ffff0000ffffll,\n"
"    0x00000000ffffffffll\n"
"  };\n"
"  uint64_t A;\n"
"  A= X ^ Y;\n"
"  \n"
"  for(i=1;i<64;i<<=1) A |= (A<<i);\n"
"  A = ~A;\n"
"  for(j=0,i=1;i<64;i<<=1,j++) A= (A&B[j]) + ((A>>i)&B[j]);\n"
"  return A;\n"
"}\n"
"\n"
"\n"
"static void* qumap_find(qumap_t *map,uint64_t key)\n"
"{\n"
"  qumap_node_t *node;\n"
"\n"
"  node= map->root;\n"
"  if (!node) return NULL;\n"
"\n"
"  while(node->internal) node= _QUMAP_CHLD(node, qumap_dir(key,node));\n"
"\n"
"  if (_QUMAP_KEY(node)==key) return _QUMAP_DATA(node);\n"
"                          else return NULL;\n"
"}\n"
"\n"
"\n"
"\n"
"static void* qumap_insert(qumap_t* map, uint64_t key, void *data)\n"
"{\n"
"  qumap_node_t *node_elt;\n"
"\n"
"  if (!map->root)\n"
"  {\n"
"     map->root= qumap_leaf_node(key,data);\n"
"     return 0;\n"
"  }\n"
"\n"
"  node_elt= map->root;\n"
"  while(node_elt->internal)\n"
"    node_elt= _QUMAP_CHLD(node_elt, qumap_dir(key,node_elt));\n"
"\n"
"  \n"
"  uint64_t elt;\n"
"  int dir_elt;\n"
"  uint8_t bit;\n"
"  qumap_node_t *node_key;\n"
"\n"
"  elt= _QUMAP_KEY(node_elt);\n"
"\n"
"  if (elt==key) return _QUMAP_DATA(node_elt);\n"
"\n"
"  bit= qumap_crit_bit(elt,key);\n"
"\n"
"  dir_elt= (elt>>bit)&1;\n"
"\n"
"  node_key= qumap_internal_node(bit);\n"
"  _QUMAP_CHLD(node_key,1-dir_elt)= qumap_leaf_node(key,data);\n"
"\n"
"  qumap_node_t *parent;\n"
"  qumap_node_t *node;\n"
"\n"
"  parent= 0;\n"
"  node= map->root;\n"
"\n"
"  while(node->internal)\n"
"  {\n"
"     if (node->bit>bit) break;\n"
"     parent= node;\n"
"     node= _QUMAP_CHLD(node, qumap_dir(key,node));\n"
"  }\n"
"\n"
"  _QUMAP_CHLD(node_key,dir_elt)= node;\n"
"  if (parent) { if (_QUMAP_CHLD(parent,0)==node)  _QUMAP_CHLD(parent,0)= node_key;\n"
"                                       else  _QUMAP_CHLD(parent,1)= node_key; }\n"
"  else map->root= node_key;\n"
"  return 0;\n"
"}\n"
"\n"
"static void *qumap_remove(qumap_t* map, uint64_t key)\n"
"{\n"
"  void *retv;\n"
"\n"
"  if (!map->root) return NULL;\n"
"\n"
"   \n"
"  qumap_node_t *node;\n"
"  qumap_node_t *sibling;\n"
"  qumap_node_t *parent;\n"
"  qumap_node_t *gparent;\n"
"\n"
"  node= map->root;\n"
"  gparent= 0;\n"
"  sibling= 0;\n"
"  parent= 0;\n"
"\n"
"  while(node->internal)\n"
"  {\n"
"    int dir;\n"
"    dir= qumap_dir(key, node);\n"
"    gparent= parent;\n"
"    parent= node;\n"
"    sibling= _QUMAP_CHLD(node, 1-dir);\n"
"    node= _QUMAP_CHLD(node, dir);\n"
"  }\n"
"\n"
"  if (_QUMAP_KEY(node)!=key) return NULL;\n"
"  retv= _QUMAP_DATA(node);\n"
"  free(node);\n"
"\n"
"  if (!gparent)\n"
"  {\n"
"     map->root= sibling;\n"
"     if (parent) free(parent);\n"
"     return retv;\n"
"  }\n"
"\n"
"  if (_QUMAP_CHLD(gparent,0)==parent) _QUMAP_CHLD(gparent,0)= sibling;\n"
"                            else _QUMAP_CHLD(gparent,1)= sibling;\n"
"  free(parent);\n"
"\n"
"  return retv;\n"
"}\n"
"\n"
"static void *qumap_destroy(qumap_t *map)\n"
"{\n"
"  qumap_node_t *node;\n"
"  qumap_node_t *parent;\n"
"  qumap_node_t *gparent;\n"
"  qumap_node_t *sibling;\n"
"  void *retv;\n"
"\n"
"  if (!map->root) return 0;\n"
"\n"
"  parent= 0;\n"
"  gparent= 0;\n"
"  sibling= 0;\n"
"  node= map->root;\n"
"  while(node->internal)\n"
"  {\n"
"    gparent= parent;\n"
"    parent= node;\n"
"    sibling= _QUMAP_CHLD(node, 1);\n"
"    node= _QUMAP_CHLD(node,0);\n"
"  }\n"
"  retv= _QUMAP_DATA(node);\n"
"  free(node);\n"
"  if (!gparent) \n"
"  {\n"
"    map->root= sibling;\n"
"    if (parent) free(parent);\n"
"    return retv;\n"
"  }\n"
"  if (_QUMAP_CHLD(gparent,0)==parent)\n"
"    _QUMAP_CHLD(gparent,0)= sibling;\n"
"  else\n"
"    _QUMAP_CHLD(gparent,1)= sibling;\n"
"\n"
"  free(parent);\n"
"\n"
"  return retv;\n"
"}\n"
"\n"
"static qumap_t  *qumap_new()\n"
"{\n"
"  qumap_t *s;\n"
"  s= malloc(sizeof(*s));\n"
"  s->root= NULL;\n"
"  return s;\n"
"}\n"
"\n"
"static void qumap_free(qumap_t *map)\n"
"{\n"
"  free(map);\n"
"}\n"
"\n"
"\n"
"static void flolay_get(char *f_w, char *f_r, char *d_w, char *d_r);\n"
"\n"
"\n"
"static uint8_t flolay_spc[32], *flolay_ali16;\n"
"\n"
"\n"
"static void flolay_align()\n"
"{\n"
"  uint64_t T;\n"
"  T= (uint64_t) flolay_spc;\n"
"  if (T%16) { flolay_ali16= flolay_spc + (16-T%16); }\n"
"       else { flolay_ali16= flolay_spc; }\n"
"}\n"
"\n"
"static int flolay_find_byte(int nb, uint8_t V)\n"
"{\n"
"  int i;\n"
"  for(i=0;i<nb;i++)\n"
"   if (flolay_ali16[i]==V) return i;\n"
"  return 0;\n"
"}\n"
"\n"
"static void flolay_get_flo(float *F, uint8_t *fromNative,uint8_t *toNative)\n"
"{\n"
"\n"
"  int i;\n"
"  *F= 1;\n"
"  for(i=0;i<149;i++) *F/= 2;\n"
"  for(i=0;i<3;i++) \n"
"  {\n"
"    *(float*) flolay_ali16= *F;\n"
"    toNative[i]= flolay_find_byte(4, 1);\n"
"\n"
"    *F *= 256;\n"
"  }\n"
"  *F= -0.0;\n"
"  *(float*) flolay_ali16= *F;\n"
"\n"
"  toNative[i]= flolay_find_byte(4, 0x80); \n"
"\n"
"  for(i=0;i<4;i++) fromNative[toNative[i]]= i;\n"
"}\n"
"\n"
"static void flolay_get_dou(double *F, uint8_t *fromNative,uint8_t *toNative)\n"
"{\n"
"\n"
"  int i;\n"
"  *F= 1;\n"
"  for(i=0;i<1074;i++) *F/= 2;\n"
"  for(i=0;i<7;i++) \n"
"  {\n"
"    *(double*) flolay_ali16= *F;\n"
"    toNative[i]= flolay_find_byte(8, 1);\n"
"\n"
"    *F *= 256;\n"
"  }\n"
"  *F= -0.0;\n"
"  *(double*) flolay_ali16= *F;\n"
"\n"
"  toNative[i]= flolay_find_byte(8, 0x80); \n"
"  for(i=0;i<8;i++) fromNative[toNative[i]]= i;\n"
"}\n"
"\n"
"\n"
"static void flolay_get(char *f_w, char *f_r, char *d_w, char *d_r)\n"
"{\n"
"  float F;\n"
"  double D;\n"
"\n"
"  flolay_align();\n"
"  flolay_get_flo(&F, (uint8_t*) f_w, (uint8_t*) f_r);\n"
"  flolay_get_dou(&D, (uint8_t*) d_w, (uint8_t*) d_r);\n"
"}\n"
"\n"
"\n"
"struct _bff\n"
"{\n"
"  fbq_t *free_block_queue;\n"
"  fbm_t *free_block_map;\n"
"  qumap_t *object_map;\n"
"  uint64_t file_size;\n"
"  uint8_t magic[16];\n"
"  uint64_t next_avail_ident;\n"
"\n"
"  uint8_t *buffer;\n"
"  uint64_t buffer_size;\n"
"\n"
"  int mode;\n"
"  char *file_name;\n"
"  int file;\n"
"  int have_error;\n"
"  int eno;\n"
"#ifdef DEBUG\n"
"  uint64_t f_pos, f_buffer_size;\n"
"  uint8_t *f_buffer;\n"
"#endif\n"
"\n"
"  uint32_t o_type;\n"
"  uint64_t o_size, o_len;\n"
"  uint8_t *o_buffer, *o_end, *o_pos;\n"
"\n"
"  jmp_buf ejump;\n"
"  int n_ralloc, s_ralloc;\n"
"  void **ralloc;\n"
"\n"
"  int w_pass;\n"
"\n"
"  postpone_buffer_t *pb;\n"
"\n"
"  uint8_t  floio_store[32];\n"
"  uint8_t *floio_cptr;\n"
"  float   *floio_fptr;\n"
"  double  *floio_dptr;\n"
"  char     floio_perm_fw[4], floio_perm_fr[4],\n"
"           floio_perm_dw[8], floio_perm_dr[8];\n"
"};\n"
"\n"
"\n"
"static int bff_read_file_header(bff_t *B);\n"
"static int bff_write_file_header(bff_t *B);\n"
"static int bff_write_block_header\n"
"    (bff_t *bf, uint64_t size, uint64_t prev, uint64_t next);\n"
"static int bff_read_block_header\n"
"    (bff_t *bf, uint64_t *size, uint64_t *prev, uint64_t *next);\n"
"static int bff_write_object_header\n"
"    (bff_t *bf, uint64_t size);\n"
"static int bff_read_object_header\n"
"    (bff_t *bf, uint64_t *size);\n"
"static int bff_write_object\n"
"   (bff_t *bf, uint64_t ident, uint64_t type, uint64_t osize, uint8_t *odata);\n"
"static int bff_write_object_eof\n"
"   (bff_t *bf, object_descriptor_t *O,uint64_t osize,  uint8_t *odata);\n"
"static int bff_write_object_single_free_block\n"
"       (bff_t *bf, object_descriptor_t *O,uint64_t osize,  uint8_t *odata, free_block_t *A);\n"
"static int bff_write_object_two_free_blocks\n"
"       (bff_t *bf, object_descriptor_t *O, uint64_t osize, uint8_t *odata,\n"
"        free_block_t *A, free_block_t *B);\n"
"static int bff_write_object_free_block_plus_eof\n"
"   (bff_t *bf, object_descriptor_t *O, \n"
"    uint64_t osize, uint8_t *odata, free_block_t *A);\n"
"static int bff_free_object\n"
"   (bff_t *bf, uint64_t ident);\n"
"static void bff_record_free_block\n"
"       (bff_t *bf, free_block_t *F);\n"
"static void bff_record_object\n"
"       (bff_t *bf, object_descriptor_t *O);\n"
"static int _bff_create(const char *fn,const uint8_t magic[], bff_t **R);\n"
"static int _bff_open\n"
"    (const char *fn, int mode, bff_t **R, const uint8_t magic[]);\n"
"static void _bff_close(bff_t *bf);\n"
"static bff_t *bff_new();\n"
"static int bff_write_bytes(bff_t *bf, uint64_t size, uint8_t *data);\n"
"static int bff_read_bytes(bff_t *bf, uint64_t size, uint8_t *data);\n"
"static int bff_write_formatted(bff_t *bf, const char *fmt, ...);\n"
"static int bff_read_formatted(bff_t *bf, const char *fmt, ...);\n"
"static int bff_seek(bff_t *bf, uint64_t pos);\n"
"static int bff_trunc(bff_t *bf, uint64_t size);\n"
"static int bff_write_junk(bff_t *bf, uint64_t size);\n"
"static int bff_read_index(bff_t *bf);\n"
"static void bff_index_entry\n"
"    (bff_t *bf, uint32_t type, uint64_t offset, uint64_t ident);\n"
"static int decode_unt32(uint8_t *d, uint32_t *R);\n"
"static int decode_unt64(uint8_t *d, uint64_t *R);\n"
"static int encode_unt64(uint8_t *d, uint64_t V);\n"
"static int encode_unt32(uint8_t *d, uint32_t V);\n"
"static int bff_write_index(bff_t *bf);\n"
"static int bff_read_object(bff_t *bf, uint64_t ident);\n"
"static void bff_fail_bread(bff_t *bf);\n"
"static int bff_bread_u8(bff_t *bf, uint8_t *R);\n"
"static int bff_bread_i8(bff_t *bf, int8_t *R) \n"
"   ;\n"
"static int bff_bread_u16(bff_t *bf, uint16_t *R);\n"
"static int bff_bread_i16(bff_t *bf, int16_t *R) \n"
"   ;\n"
"static int bff_bread_u32(bff_t *bf, uint32_t *R);\n"
"static int bff_bread_i32(bff_t *bf, int32_t *R) \n"
"   ;\n"
"static int bff_bread_u64(bff_t *bf, uint64_t *R);\n"
"static int bff_bread_i64(bff_t *bf, int64_t *R) \n"
"   ;\n"
"static int bff_bread_s8(bff_t *bf, uint8_t **R);\n"
"static int bff_bread_s16(bff_t *bf, uint16_t **R);\n"
"static int bff_bread_s32(bff_t *bf, uint32_t **R);\n"
"static void *bff_ralloc(bff_t *bf, uint32_t size);\n"
"static void *bff_ralloc0(bff_t *bf, uint32_t size);\n"
"static void bff_free_ralloc(bff_t *bf);\n"
"static void bff_reset_ralloc(bff_t *bf);\n"
"static void bff_bwrite_u8(bff_t *bf, uint8_t V);\n"
"static void bff_bwrite_u16(bff_t *bf, uint16_t V);\n"
"static void bff_bwrite_u32(bff_t *bf, uint32_t V);\n"
"static void bff_bwrite_u64(bff_t *bf, uint64_t V);\n"
"static void bff_bwrite_i8(bff_t *bf, int8_t V)\n"
"   ;\n"
"static void bff_bwrite_i16(bff_t *bf, int16_t V)\n"
"   ;\n"
"static void bff_bwrite_i32(bff_t *bf, int32_t V)\n"
"   ;\n"
"static void bff_bwrite_i64(bff_t *bf, int64_t V)\n"
"   ;\n"
"static void bff_bwrite_s8(bff_t *bf, uint8_t *D);\n"
"static void bff_bwrite_s16(bff_t *bf, uint16_t *D);\n"
"static void bff_bwrite_s32(bff_t *bf, uint32_t *D);\n"
"static int str16len(uint16_t *S);\n"
"static int str32len(uint32_t *S);\n"
"static void bff_alloc_o_buffer(bff_t *bf, uint64_t size);\n"
"static int bff_encode_and_write\n"
"   (bff_t *bf, void *obj, uint32_t otype,\n"
"    void (*encode)(bff_t *,void*), uint64_t *objidptr);\n"
"static int bff_continue_postponed_writes(bff_t *bf);\n"
"static void bff_pointercheck\n"
"   (bff_t *bf, void *obj, uint64_t *objid, int (*writefunc)(void *, void*));\n"
"static void bff_init_floio(bff_t *bf);\n"
"static int bff_bread_f32(bff_t *bf, float *R);\n"
"static void bff_bwrite_f32(bff_t *bf, float V);\n"
"static int bff_bread_f64(bff_t *bf, double *R);\n"
"static void bff_bwrite_f64(bff_t *bf, double V);\n"
"\n"
"static int bff_read_file_header(bff_t *B)\n"
"{\n"
"  if (bff_seek(B,0)) return 1;\n"
"     return bff_read_formatted(B, \042bii\042, 16, \n"
"              B->magic,&B->file_size, &B->next_avail_ident);\n"
"}\n"
"\n"
"static int bff_write_file_header(bff_t *B)\n"
"{\n"
"  if (bff_seek(B,0)) return 1;\n"
"  return bff_write_formatted(B,\042bii\042,16,\n"
"     B->magic, B->file_size, B->next_avail_ident);\n"
"}\n"
"\n"
"static int bff_write_block_header\n"
"    (bff_t *bf, uint64_t size, uint64_t prev, uint64_t next)\n"
"{\n"
"  return bff_write_formatted(bf, \042iii\042, size, prev, next);\n"
"}\n"
"\n"
"static int bff_read_block_header\n"
"    (bff_t *bf, uint64_t *size, uint64_t *prev, uint64_t *next)\n"
"{\n"
"  return bff_read_formatted(bf, \042iii\042, size, prev, next);\n"
"}\n"
"\n"
"static int bff_write_object_header\n"
"    (bff_t *bf, uint64_t size)\n"
"{\n"
"  return bff_write_formatted(bf, \042i\042, size);\n"
"}\n"
"\n"
"static int bff_read_object_header\n"
"    (bff_t *bf, uint64_t *size)\n"
"{\n"
"  return bff_read_formatted(bf, \042i\042, size);\n"
"}\n"
"\n"
"static int bff_write_object\n"
"   (bff_t *bf, uint64_t ident, uint64_t type, uint64_t osize, uint8_t *odata)\n"
"{\n"
"  free_block_t *A, *B;\n"
"  uint64_t rsize;\n"
"  object_descriptor_t *O;\n"
"\n"
"  O= malloc(sizeof(*O));\n"
"  O->type= type;\n"
"  O->ident= ident;\n"
"\n"
"  A= fbq_get(bf->free_block_queue);\n"
"  if (!A) \n"
"    return bff_write_object_eof(bf, O, osize, odata);\n"
"\n"
"  rsize= osize+BLOCK_HEADER_SIZE+OBJECT_HEADER_SIZE;\n"
"  fbm_remove(bf->free_block_map, A->offset);\n"
"  if (A->size>=rsize)\n"
"    return bff_write_object_single_free_block(bf, O, osize, odata, A);\n"
"\n"
"  rsize= osize+2*BLOCK_HEADER_SIZE+OBJECT_HEADER_SIZE;\n"
"  B= fbq_peek(bf->free_block_queue);\n"
"  if (!B || (A->size+B->size<rsize))\n"
"    return bff_write_object_free_block_plus_eof(bf, O, osize, odata,A);\n"
"\n"
"  B= fbq_get(bf->free_block_queue);\n"
"  fbm_remove(bf->free_block_map, B->offset);\n"
"  return bff_write_object_two_free_blocks(bf, O, osize, odata,A,B);\n"
"}\n"
"\n"
"static int bff_write_object_eof\n"
"   (bff_t *bf, object_descriptor_t *O,uint64_t osize,  uint8_t *odata)\n"
"{\n"
"  uint64_t bsize, rsize;\n"
"  O->offset= bf->file_size;\n"
"  bsize= rsize= osize + BLOCK_HEADER_SIZE + OBJECT_HEADER_SIZE; \n"
"  if (bsize<FREE_BLOCK_THRESHOLD)\n"
"     bsize= FREE_BLOCK_THRESHOLD;\n"
"  if (\n"
"    bff_seek(bf,bf->file_size) || \n"
"    bff_write_block_header(bf, bsize, 0, 0) ||\n"
"    bff_write_object_header(bf, osize) ||\n"
"    bff_write_bytes(bf, osize, odata) ||\n"
"    bff_write_junk(bf, bsize-rsize) \n"
"  )\n"
"     return 1;\n"
"  bf->file_size+= bsize;\n"
"  bff_record_object(bf, O);\n"
"  return 0;\n"
"}\n"
"\n"
"static int bff_write_object_single_free_block\n"
"       (bff_t *bf, object_descriptor_t *O,uint64_t osize,  uint8_t *odata, free_block_t *A)\n"
"{\n"
"  uint64_t rsize,bsize;\n"
"  if (bff_seek(bf,A->offset)) return 1;\n"
"  O->offset= A->offset;\n"
"  rsize= osize + OBJECT_HEADER_SIZE + BLOCK_HEADER_SIZE;\n"
"  if (A->size-rsize<FREE_BLOCK_THRESHOLD)\n"
"    bsize= A->size; \n"
"  else\n"
"    bsize= rsize;\n"
"  if (bff_write_block_header(bf, bsize, 0, 0) ||\n"
"      bff_write_object_header(bf, osize) ||\n"
"      bff_write_bytes(bf, osize, odata)) return 1;\n"
"  A->offset+= bsize;\n"
"  A->size-= bsize;\n"
"  if (A->size==0) \n"
"     free(A);\n"
"  else\n"
"     bff_record_free_block(bf, A);\n"
"  bff_record_object(bf, O);\n"
"  return 0;\n"
"}\n"
"\n"
"static int bff_write_object_two_free_blocks\n"
"       (bff_t *bf, object_descriptor_t *O, uint64_t osize, uint8_t *odata,\n"
"        free_block_t *A, free_block_t *B)\n"
"{\n"
"  // B is smaller than A. I\047ll use it completely, than put some\n"
"  // more into A. These variables record how much of the object\n"
"  // is put into each free block\n"
"  uint64_t osize_B, osize_A;\n"
"  // block sizes for each block\n"
"  uint64_t bsize_A, bsize_B;\n"
"\n"
"  // these are the values to be used in sequence\n"
"  uint64_t OSIZ[2], bsize[2], offset[2];\n"
"\n"
"  if (B->offset<A->offset)\n"
"     osize_B= B->size - OBJECT_HEADER_SIZE-BLOCK_HEADER_SIZE;\n"
"  else\n"
"     osize_B= B->size - BLOCK_HEADER_SIZE;\n"
"\n"
"  if (osize_B>osize) osize_B= osize;\n"
"\n"
"  osize_A= osize-osize_B;\n"
"\n"
"  if (A->offset<B->offset)\n"
"    bsize_A= osize_A + OBJECT_HEADER_SIZE + BLOCK_HEADER_SIZE;\n"
"  else\n"
"    bsize_A= osize_A + BLOCK_HEADER_SIZE;\n"
"\n"
"  if (A->size-bsize_A<FREE_BLOCK_THRESHOLD)\n"
"  {\n"
"     if (A->offset<B->offset)\n"
"     {\n"
"       osize_A+= A->size-bsize_A;\n"
"       osize_B-= A->size-bsize_A;\n"
"     }\n"
"     bsize_A= A->size;\n"
"  } \n"
"  bsize_B= B->size;\n"
"\n"
"  if (A->offset<B->offset)\n"
"  {\n"
"    OSIZ[0]= osize_A;\n"
"    bsize[0]= bsize_A;\n"
"    offset[0]= A->offset;\n"
"    OSIZ[1]= osize_B;\n"
"    bsize[1]= bsize_B;\n"
"    offset[1]= B->offset;\n"
"  }\n"
"  else\n"
"  {\n"
"    OSIZ[0]= osize_B;\n"
"    bsize[0]= bsize_B;\n"
"    offset[0]= B->offset;\n"
"    OSIZ[1]= osize_A;\n"
"    bsize[1]= bsize_A;\n"
"    offset[1]= A->offset;\n"
"  }\n"
"\n"
"  if (\n"
"    bff_seek(bf, offset[0]) ||\n"
"    bff_write_block_header(bf, bsize[0], 0, offset[1]) ||\n"
"    bff_write_object_header(bf, osize) ||\n"
"    bff_write_bytes(bf, OSIZ[0], odata) || \n"
"    bff_seek(bf, offset[1]) ||\n"
"    bff_write_block_header(bf, bsize[1], offset[0], 0) ||\n"
"    bff_write_bytes(bf, OSIZ[1], odata+OSIZ[0])\n"
"  )\n"
"      return 1;\n"
"\n"
"  free(B);\n"
"  A->offset+= bsize_A;\n"
"  A->size-= bsize_A;\n"
"  if (A->size==0)\n"
"    free(A); \n"
"  else\n"
"    bff_record_free_block(bf,A);\n"
"  O->offset= offset[0];\n"
"  bff_record_object(bf, O);\n"
"  return 0;\n"
"}\n"
"\n"
"static int bff_write_object_free_block_plus_eof\n"
"   (bff_t *bf, object_descriptor_t *O, \n"
"    uint64_t osize, uint8_t *odata, free_block_t *A)\n"
"{\n"
"  uint64_t offset2, offset1;\n"
"  uint64_t osize_A, osize_EOF;\n"
"  uint64_t bsize_A, bsize_EOF, rsize_EOF;\n"
"\n"
"  osize_A= A->size - BLOCK_HEADER_SIZE - OBJECT_HEADER_SIZE;\n"
"  osize_EOF= osize - osize_A;\n"
"\n"
"  bsize_A= A->size;\n"
"  rsize_EOF= bsize_EOF= osize_EOF + BLOCK_HEADER_SIZE;\n"
"\n"
"  if (bsize_EOF < FREE_BLOCK_THRESHOLD)\n"
"    bsize_EOF= FREE_BLOCK_THRESHOLD;\n"
"\n"
"  offset1= A->offset;\n"
"  offset2= bf->file_size;\n"
"\n"
"  free(A);\n"
"  if (\n"
"    bff_seek(bf, offset1) ||\n"
"    bff_write_block_header(bf, bsize_A, 0, offset2) ||\n"
"    bff_write_object_header(bf, osize) ||\n"
"    bff_write_bytes(bf, osize_A, odata) ||\n"
"    bff_seek(bf, offset2) ||\n"
"    bff_write_block_header(bf, bsize_EOF, offset1, 0) ||\n"
"    bff_write_bytes(bf, osize_EOF, odata+osize_A) ||\n"
"    bff_write_junk(bf, bsize_EOF - rsize_EOF)\n"
"  )\n"
"     return 1;\n"
"\n"
"  bf->file_size+= bsize_EOF;\n"
"  O->offset= offset1;\n"
"  bff_record_object(bf, O);\n"
"  return 0;\n"
"}\n"
"\n"
"static int bff_free_object\n"
"   (bff_t *bf, uint64_t ident)\n"
"{\n"
"  free_block_t *F;\n"
"  uint64_t prev, next, size;\n"
"  object_descriptor_t *O;\n"
"  uint64_t bo;\n"
"\n"
"  O= qumap_remove(bf->object_map, ident);\n"
"  if (!O) return 1;\n"
"\n"
"  bo= O->offset;\n"
"  free(O);\n"
"\n"
"  while(1)\n"
"  {\n"
"    if (bff_seek(bf, bo) ||\n"
"        bff_read_block_header(bf, &size, &prev, &next)) \n"
"      return 1;\n"
"    F= calloc(1,sizeof(*F));\n"
"    F->offset= bo;\n"
"    F->size= size;\n"
"    bff_record_free_block(bf,F);\n"
"    bo= next;\n"
"    if (!bo) break;\n"
"  }\n"
"\n"
"  return 0;\n"
"}\n"
"\n"
"static void bff_record_free_block\n"
"       (bff_t *bf, free_block_t *F)\n"
"{\n"
"  free_block_t *A,*B;\n"
"\n"
"  A= fbm_remove(bf->free_block_map, F->offset-1);\n"
"  B= fbm_remove(bf->free_block_map, F->offset+F->size);\n"
"\n"
"  if (A) { F->offset= A->offset; F->size+= A->size;\n"
"           fbq_remove(bf->free_block_queue,A); free(A); }\n"
"  if (B) { F->size+= B->size; \n"
"           fbq_remove(bf->free_block_queue, B); free(B); }\n"
"\n"
"  if (F->offset+F->size==bf->file_size)\n"
"  {\n"
"    bf->file_size= F->offset;\n"
"    free(F);\n"
"    return ;\n"
"  }\n"
"  fbm_insert(bf->free_block_map, F);\n"
"  fbq_put(bf->free_block_queue, F);\n"
"}\n"
"\n"
"static void bff_record_object\n"
"       (bff_t *bf, object_descriptor_t *O)\n"
"{\n"
"  qumap_insert(bf->object_map, O->ident, O);\n"
"}\n"
"\n"
"\n"
"static int _bff_create(const char *fn,const uint8_t magic[], bff_t **R)\n"
"{\n"
"  bff_t *bf;\n"
"  int retv;\n"
"  retv= 0;\n"
"  bf= bff_new();\n"
"  bf->file_name= strdup(fn);\n"
"  memcpy(bf->magic, magic, 16);\n"
"  bf->file= open(fn, O_RDWR | O_CREAT | O_TRUNC, 0600);\n"
"  if (bf->file==-1) \n"
"  {\n"
"    bf->have_error= 1;\n"
"    bf->eno= errno;\n"
"    retv= 1;\n"
"  }\n"
"  bf->mode= \047W\047;\n"
"  bff_write_file_header(bf);\n"
"  bff_write_junk(bf, bf->file_size - FILE_HEADER_SIZE);\n"
"  *R= bf;\n"
"  return retv;\n"
"}\n"
"\n"
"static int _bff_open\n"
"    (const char *fn, int mode, bff_t **R, const uint8_t magic[])\n"
"{\n"
"  bff_t *bf;\n"
"  int i, retv;\n"
"  retv= 0;\n"
"  bf= bff_new();\n"
"  bf->file_name= strdup(fn);\n"
"  bf->file= open(fn, O_RDWR);\n"
"  if (bf->file==-1) \n"
"  {\n"
"    bf->have_error= 1;\n"
"    bf->eno= errno;\n"
"    retv= 1;\n"
"  }\n"
"  bf->mode= mode;\n"
"  if (bf->mode) bf->mode=\047W\047; else bf->mode= \047R\047;\n"
"  bff_read_file_header(bf);\n"
"  for(i=0;i<16;i++) \n"
"    if (bf->magic[i]!=magic[i]) \n"
"    {\n"
"      bf->have_error= 1;\n"
"      bf->eno= EBADF;\n"
"      retv= 1;\n"
"      break;\n"
"    }\n"
"  if (!retv)\n"
"    bff_read_index(bf);\n"
"  *R= bf;\n"
"  return retv;\n"
"}\n"
"\n"
"static void _bff_close(bff_t *bf)\n"
"{\n"
"  object_descriptor_t *O;\n"
"  free_block_t *F;\n"
"  if (bf->mode==\047W\047)\n"
"  {\n"
"    bff_write_index(bf);\n"
"    bff_write_file_header(bf);\n"
"  }\n"
"\n"
"  while((F=fbm_destroy(bf->free_block_map))) ;\n"
"  while((F=fbq_get(bf->free_block_queue)))\n"
"    free(F);\n"
"  fbq_free(bf->free_block_queue);\n"
"  while((O=qumap_destroy(bf->object_map)))\n"
"    free(O);\n"
"  qumap_free(bf->object_map);\n"
"  \n"
"  if (bf->buffer) free(bf->buffer);\n"
"  if (bf->file_name) free(bf->file_name);\n"
"  if (bf->file>=0) close(bf->file);\n"
"  bff_free_ralloc(bf);\n"
"  if (bf->ralloc) free(bf->ralloc);\n"
"  memset(bf, 0, sizeof(*bf));\n"
"  free(bf);\n"
"}\n"
"\n"
"static bff_t *bff_new()\n"
"{\n"
"  bff_t *bf;\n"
"  bf= calloc(1,sizeof(*bf));\n"
"  bf->free_block_queue= fbq_new(100);\n"
"  bf->free_block_map= fbm_new();\n"
"  bf->object_map= qumap_new();\n"
"  bf->buffer_size= 1000;\n"
"  bf->buffer= malloc(bf->buffer_size);\n"
"\n"
"  bf->next_avail_ident= 32;\n"
"\n"
"  // reserve this for the file header\n"
"  bf->file_size= 128;\n"
"#ifdef DEBUG\n"
"  bf->f_buffer_size= 1000;\n"
"  bf->f_buffer= malloc(bf->f_buffer_size);\n"
"  bf->f_pos= 0;\n"
"#endif\n"
"  bf->pb= postpone_buffer_new(50);\n"
"  bff_init_floio(bf);\n"
"  return bf;\n"
"}\n"
"\n"
"static int bff_write_bytes(bff_t *bf, uint64_t size, uint8_t *data)\n"
"{\n"
"  if (!size) return 0;\n"
"#ifdef DEBUG\n"
"  if (bf->f_pos+size>bf->f_buffer_size)\n"
"  {\n"
"    bf->f_buffer_size+= bf->f_buffer_size/2+size;\n"
"    bf->f_buffer= realloc(bf->f_buffer, bf->f_buffer_size);\n"
"  }\n"
"  memcpy(bf->f_buffer+bf->f_pos, data, size);\n"
"  bf->f_pos+= size;\n"
"  return 0;\n"
"#endif\n"
"  if (bf->file<0 || bf->mode==\047R\047 || bf->have_error) return 1;\n"
"  ssize_t K;\n"
"  while(size)\n"
"  {\n"
"    K= write(bf->file, data, size);\n"
"    if (K<=0)\n"
"    {\n"
"      if (errno==EINTR) continue;\n"
"      bf->eno= errno;\n"
"      bf->have_error= 1;\n"
"      return 1;\n"
"    }\n"
"    data+= K;\n"
"    size-= K;\n"
"  }\n"
"  return 0;\n"
"}\n"
"\n"
"static int bff_read_bytes(bff_t *bf, uint64_t size, uint8_t *data)\n"
"{\n"
"  if (size==0) return 0;\n"
"#ifdef DEBUG\n"
"  if (bf->f_pos+size>bf->f_buffer_size) return 1;\n"
"  memcpy(data, bf->f_buffer+bf->f_pos, size);\n"
"  bf->f_pos+= size;\n"
"  return 0;\n"
"#endif\n"
"  if (bf->file<0 || bf->have_error) return 1;\n"
"  ssize_t K;\n"
"  while(size)\n"
"  {\n"
"    K= read(bf->file, data, size);\n"
"    if (K<=0)\n"
"    {\n"
"      if (errno==EINTR) continue;\n"
"      bf->eno= errno;\n"
"      bf->have_error= 1;\n"
"      return 1;\n"
"    }\n"
"    data+= K;\n"
"    size-= K;\n"
"  }\n"
"  return 0;\n"
"}\n"
"\n"
"static int bff_write_formatted(bff_t *bf, const char *fmt, ...)\n"
"{\n"
"  va_list ap;\n"
"  int sz;\n"
"  uint8_t *k; uint64_t M;\n"
"  int i,j,b;\n"
"  va_start(ap,fmt);\n"
"  sz= 0;\n"
"  for(i=0;fmt[i];i++)\n"
"   switch(fmt[i])\n"
"   { case \047b\047: sz+= va_arg(ap, int); break;\n"
"     case \047i\047: sz+= 8; va_arg(ap, uint64_t); break; }\n"
"  va_end(ap);\n"
"  if (bf->buffer_size<sz) \n"
"  {\n"
"    bf->buffer_size= sz;\n"
"    bf->buffer= realloc(bf->buffer, bf->buffer_size);\n"
"  }\n"
"  sz= 0;\n"
"  va_start(ap,fmt);\n"
"  for(i=0;fmt[i];i++)\n"
"   switch(fmt[i])\n"
"   { case \047b\047: b= va_arg(ap, int); k= va_arg(ap, uint8_t*);\n"
"               memcpy(bf->buffer+sz, k, b);\n"
"               sz+= b;\n"
"               break;\n"
"     case \047i\047: M= va_arg(ap, uint64_t);\n"
"               for(j=0;j<8;j++) { bf->buffer[sz+j]= M; M>>= 8; }\n"
"               sz+= 8;\n"
"               break; }\n"
"  va_end(ap);\n"
"  return bff_write_bytes(bf, sz, bf->buffer);\n"
"}\n"
"\n"
"static int bff_read_formatted(bff_t *bf, const char *fmt, ...)\n"
"{\n"
"  va_list ap;\n"
"  int sz;\n"
"  uint8_t *k; uint64_t *M,V,U;\n"
"  int i,j,b;\n"
"  va_start(ap,fmt);\n"
"  sz= 0;\n"
"  for(i=0;fmt[i];i++)\n"
"   switch(fmt[i])\n"
"   { case \047b\047: sz+= va_arg(ap, int); break;\n"
"     case \047i\047: sz+= 8; va_arg(ap, uint64_t); break; }\n"
"  va_end(ap);\n"
"  if (bf->buffer_size<sz) \n"
"  {\n"
"    bf->buffer_size= sz;\n"
"    bf->buffer= realloc(bf->buffer, bf->buffer_size);\n"
"  }\n"
"  if (bff_read_bytes(bf, sz, bf->buffer)) return 1;\n"
"  sz= 0;\n"
"  va_start(ap,fmt);\n"
"  for(i=0;fmt[i];i++)\n"
"   switch(fmt[i])\n"
"   { case \047b\047: b= va_arg(ap, int); k= va_arg(ap, uint8_t*);\n"
"               memcpy(k, bf->buffer+sz, b);\n"
"               sz+= b;\n"
"               break;\n"
"     case \047i\047: M= va_arg(ap, uint64_t*);\n"
"               V= 0; \n"
"               for(j=0;j<8;j++) { U= bf->buffer[sz+j]; V|= U<<(j*8); }\n"
"               *M= V;\n"
"               sz+= 8;\n"
"               break; }\n"
"  va_end(ap);\n"
"  return 0;\n"
"}\n"
"\n"
"\n"
"static int bff_seek(bff_t *bf, uint64_t pos)\n"
"{\n"
"#ifdef DEBUG\n"
"  if (pos>bf->f_buffer_size)\n"
"  {\n"
"    bf->f_buffer_size= pos + 1000;\n"
"    bf->f_buffer= realloc(bf->f_buffer, bf->f_buffer_size);\n"
"  }\n"
"  bf->f_pos= pos;\n"
"  return 0;\n"
"#endif\n"
"  if (bf->file<0 || bf->have_error) return 1;\n"
"  if (lseek(bf->file,pos,SEEK_SET)==(off_t)-1)\n"
"  {\n"
"    bf->eno= errno;\n"
"    bf->have_error= 1;\n"
"    return 1;\n"
"  }\n"
"  return 0;\n"
"}\n"
"\n"
"static int bff_trunc(bff_t *bf, uint64_t size)\n"
"{\n"
"  if (bf->have_error || bf->file<0) return 1;\n"
"  if (ftruncate(bf->file, size)) return 1;\n"
"  return 0;\n"
"}\n"
"\n"
"static int bff_write_junk(bff_t *bf, uint64_t size)\n"
"{\n"
"  static uint8_t g[200];\n"
"  uint64_t cp;\n"
"  if (size==0) return 0;\n"
"  while(size)\n"
"  {\n"
"    cp= size<sizeof(g) ? size : sizeof(g);\n"
"    if (bff_write_bytes(bf, cp, g)) return 1;\n"
"    size-= cp;\n"
"  }\n"
"  return 0;\n"
"}\n"
"\n"
"\n"
"static int bff_read_index(bff_t *bf)\n"
"{\n"
"  uint8_t data[20*100], *d;\n"
"  uint32_t i, cp;\n"
"  uint64_t N;\n"
"  uint32_t T; uint64_t O,I;\n"
"  if (bff_seek(bf, bf->file_size)) return 1;\n"
"  if (bff_read_bytes(bf, 8, data)) return 1;\n"
"  decode_unt64(data,&N);\n"
"  while(N)\n"
"  {\n"
"    cp= 100;\n"
"    if (cp>N) cp= N;\n"
"    if (bff_read_bytes(bf, cp*20, data)) return 1;\n"
"    d= data;\n"
"    for(i=0;i<cp;i++)\n"
"    {\n"
"      d+= decode_unt32(d, &T);\n"
"      d+= decode_unt64(d, &O);\n"
"      d+= decode_unt64(d, &I);\n"
"      bff_index_entry(bf, T, O, I);\n"
"    }\n"
"    N-= cp;\n"
"  }\n"
"  return 0;\n"
"}\n"
"\n"
"static void bff_index_entry\n"
"    (bff_t *bf, uint32_t type, uint64_t offset, uint64_t ident)\n"
"{\n"
"  if (type==0)\n"
"  {\n"
"    free_block_t *F;\n"
"    F= calloc(1,sizeof(*F));\n"
"    F->offset= offset;\n"
"    F->size= ident;\n"
"    fbm_insert(bf->free_block_map, F);\n"
"    fbq_put(bf->free_block_queue, F);\n"
"  } \n"
"  else\n"
"  {\n"
"    object_descriptor_t *O; \n"
"    O= calloc(1,sizeof(*O));\n"
"    O->ident= ident;\n"
"    O->type= type;\n"
"    O->offset= offset;\n"
"    qumap_insert(bf->object_map, O->ident, O);\n"
"  }\n"
"}\n"
"\n"
"static int decode_unt32(uint8_t *d, uint32_t *R)\n"
"{\n"
"  uint32_t V,L,i;\n"
"  V= 0;\n"
"  for(i=0;i<4;i++)\n"
"  {\n"
"     L= d[i];\n"
"     V|= L << (i*8);\n"
"  }\n"
"  *R= V;\n"
"  return 4;\n"
"}\n"
"\n"
"static int decode_unt64(uint8_t *d, uint64_t *R)\n"
"{\n"
"  uint64_t V,L,i;\n"
"  V= 0;\n"
"  for(i=0;i<8;i++)\n"
"  {\n"
"     L= d[i];\n"
"     V|= L << (i*8);\n"
"  }\n"
"  *R= V;\n"
"  return 8;\n"
"}\n"
"\n"
"static int encode_unt64(uint8_t *d, uint64_t V)\n"
"{\n"
"  int i;\n"
"  for(i=0;i<8;i++)\n"
"  {\n"
"    d[i]= V;\n"
"    V>>= 8;\n"
"  }\n"
"  return 8;\n"
"}\n"
"\n"
"static int encode_unt32(uint8_t *d, uint32_t V)\n"
"{\n"
"  int i;\n"
"  for(i=0;i<4;i++)\n"
"  {\n"
"    d[i]= V;\n"
"    V>>= 8;\n"
"  }\n"
"  return 4;\n"
"}\n"
"\n"
"static int bff_write_index(bff_t *bf)\n"
"{\n"
"  object_descriptor_t *O;\n"
"  free_block_t *F;\n"
"  uint64_t N;\n"
"  uint8_t d[20*100];\n"
"  uint8_t *data,*end;\n"
"\n"
"  if (bf->have_error || bf->mode==\047R\047) return 1;\n"
"  if (bff_seek(bf, bf->file_size) || bff_write_junk(bf, 8)) return 1;\n"
"\n"
"  end= d + sizeof(d);\n"
"  data= d;\n"
"  N= 0;\n"
"\n"
"  while((F=fbm_destroy(bf->free_block_map)))\n"
"  {\n"
"    data+= encode_unt32(data, 0);\n"
"    data+= encode_unt64(data, F->offset);\n"
"    data+= encode_unt64(data, F->size);\n"
"    N++;\n"
"    if (data==end)\n"
"    {\n"
"       if (bff_write_bytes(bf, data-d, d)) return 1;\n"
"       data= d;\n"
"    }\n"
"  }\n"
"\n"
"  while((O=qumap_destroy(bf->object_map)))\n"
"  {\n"
"    data+= encode_unt32(data, O->type);\n"
"    data+= encode_unt64(data, O->offset);\n"
"    data+= encode_unt64(data, O->ident);\n"
"    N++;\n"
"    free(O);\n"
"    if (data==end)\n"
"    {\n"
"       if (bff_write_bytes(bf, data-d, d)) return 1;\n"
"       data= d;\n"
"    }\n"
"  }\n"
"\n"
"  if (data!=d && bff_write_bytes(bf, data-d, d)) return 1;\n"
"  \n"
"  if (bff_seek(bf, bf->file_size) || \n"
"      bff_write_formatted(bf,\042i\042,N)) \n"
"    return 1;\n"
"\n"
"  if (bff_trunc(bf, bf->file_size + 8 + 20*N)) return 1;\n"
"  return 0;\n"
"}\n"
"\n"
"static int bff_read_object(bff_t *bf, uint64_t ident)\n"
"{\n"
"  object_descriptor_t *O;\n"
"  uint8_t *data;\n"
"  uint64_t osize, bsize, bprev, bnext, rsize, hdr_size;\n"
"\n"
"  O= qumap_find(bf->object_map, ident);\n"
"  if (!O) return 1;\n"
"  bf->o_type= O->type;\n"
"  if (bff_seek(bf, O->offset)) return 1;\n"
"  if (bff_read_block_header(bf, &bsize, &bprev, &bnext))\n"
"     return 1;\n"
"  if (bff_read_object_header(bf, &osize)) return 1;\n"
"  bf->o_len= osize;\n"
"  bff_alloc_o_buffer(bf, osize);\n"
"  bf->o_end= bf->o_buffer+ bf->o_len;\n"
"  bf->o_pos= bf->o_buffer;\n"
"  data= bf->o_buffer;\n"
"  // printf(\042object size: %d. reading \042, (int) osize);\n"
"  hdr_size= BLOCK_HEADER_SIZE+OBJECT_HEADER_SIZE;\n"
"  while(1)\n"
"  {\n"
"    rsize= bsize - hdr_size;\n"
"    if (rsize>osize) rsize= osize;\n"
"    // printf(\042 %d \042, (int) rsize);\n"
"    if (bff_read_bytes(bf, rsize, data)) return 1;\n"
"    osize-= rsize;\n"
"    data+= rsize;\n"
"    if (osize==0) break;\n"
"    if (bff_seek(bf, bnext) || \n"
"        bff_read_block_header(bf, &bsize, &bprev, &bnext)) return 1;\n"
"    hdr_size= BLOCK_HEADER_SIZE;\n"
"  }\n"
" // printf(\042\134n\042);\n"
"  return 0;\n"
"}\n"
"\n"
"\n"
"static void bff_fail_bread(bff_t *bf)\n"
"{\n"
"  bff_free_ralloc(bf);\n"
"  longjmp(bf->ejump, 1);\n"
"}\n"
"\n"
"static int bff_bread_u8(bff_t *bf, uint8_t *R)\n"
"{\n"
"  if (bf->o_pos==bf->o_end) bff_fail_bread(bf);\n"
"  *R= *bf->o_pos++;\n"
"  return 0;\n"
"}\n"
"\n"
"static int bff_bread_i8(bff_t *bf, int8_t *R) \n"
"   { return bff_bread_u8(bf, (uint8_t*) R); }\n"
"\n"
"static int bff_bread_u16(bff_t *bf, uint16_t *R)\n"
"{\n"
"  uint16_t V,D;\n"
"  if (bf->o_pos+2>bf->o_end) bff_fail_bread(bf);\n"
"  V= bf->o_pos[0];\n"
"  D= bf->o_pos[1];\n"
"  *R = V | (D<<8);\n"
"  bf->o_pos+= 2;\n"
"  return 0;\n"
"}\n"
"\n"
"static int bff_bread_i16(bff_t *bf, int16_t *R) \n"
"   { return bff_bread_u16(bf, (uint16_t*) R); }\n"
"\n"
"static int bff_bread_u32(bff_t *bf, uint32_t *R)\n"
"{\n"
"  uint32_t V,D,i;\n"
"  if (bf->o_pos+4>bf->o_end) bff_fail_bread(bf);\n"
"  V= 0;\n"
"  for(i=0;i<4;i++) { D= bf->o_pos[i]; V |= D<<(i*8); }\n"
"  *R= V;\n"
"  bf->o_pos+= 4;\n"
"  return 0;\n"
"}\n"
"\n"
"static int bff_bread_i32(bff_t *bf, int32_t *R) \n"
"   { return bff_bread_u32(bf, (uint32_t*) R); }\n"
"\n"
"\n"
"static int bff_bread_u64(bff_t *bf, uint64_t *R)\n"
"{\n"
"  uint64_t V,D,i;\n"
"  if (bf->o_pos+8>bf->o_end) bff_fail_bread(bf);\n"
"  V= 0;\n"
"  for(i=0;i<8;i++) { D= bf->o_pos[i]; V |= D<<(i*8); }\n"
"  *R= V;\n"
"  bf->o_pos+= 8;\n"
"  return 0;\n"
"}\n"
"\n"
"static int bff_bread_i64(bff_t *bf, int64_t *R) \n"
"   { return bff_bread_u64(bf, (uint64_t*) R); }\n"
"\n"
"static int bff_bread_s8(bff_t *bf, uint8_t **R)\n"
"{\n"
"  uint8_t *D;\n"
"  uint32_t N;\n"
"  if (bff_bread_u32(bf, &N)) return 1;\n"
"  D= bff_ralloc(bf,N+1);\n"
"  if (bf->o_pos+N>bf->o_end) bff_fail_bread(bf);\n"
"  memcpy(D, bf->o_pos, N);\n"
"  D[N]= 0;\n"
"  *R= D;\n"
"  bf->o_pos+= N;\n"
"  return 0;\n"
"}\n"
"\n"
"static int bff_bread_s16(bff_t *bf, uint16_t **R)\n"
"{\n"
"  uint16_t *D;\n"
"  uint32_t N,i;\n"
"  if (bff_bread_u32(bf, &N)) return 1;\n"
"  D= bff_ralloc(bf,(N+1)*sizeof(D[0]));\n"
"  if (bf->o_pos+2*N>bf->o_end) bff_fail_bread(bf);\n"
"  for(i=0;i<N;i++)\n"
"  {\n"
"    D[i]= bf->o_pos[0] | (bf->o_pos[1]<<8);\n"
"    bf->o_pos+= 2;\n"
"  }\n"
"  D[N]= 0;\n"
"  *R= D;\n"
"  return 0;\n"
"}\n"
"\n"
"static int bff_bread_s32(bff_t *bf, uint32_t **R)\n"
"{\n"
"  uint32_t *D;\n"
"  uint32_t N,i;\n"
"  if (bff_bread_u32(bf, &N)) return 1;\n"
"  D= bff_ralloc(bf,(N+1)*sizeof(D[0]));\n"
"  if (bf->o_pos+4*N>bf->o_end) bff_fail_bread(bf);\n"
"  for(i=0;i<N;i++)\n"
"  {\n"
"    D[i]= bf->o_pos[0] \n"
"          | ((uint32_t)bf->o_pos[1]<<8) \n"
"          | ((uint32_t)bf->o_pos[2]<<16) \n"
"          | ((uint32_t)bf->o_pos[3]<<24);\n"
"    bf->o_pos+= 4;\n"
"  }\n"
"  D[N]= 0;\n"
"  *R= D;\n"
"  return 0;\n"
"}\n"
"\n"
"\n"
"static void *bff_ralloc(bff_t *bf, uint32_t size)\n"
"{\n"
"  void *R;\n"
"  if (bf->n_ralloc==bf->s_ralloc)\n"
"  {\n"
"    bf->s_ralloc+= bf->s_ralloc/2 + 50;\n"
"    if (bf->ralloc) \n"
"      bf->ralloc= realloc(bf->ralloc, bf->s_ralloc*sizeof(bf->ralloc[0]));\n"
"    else\n"
"      bf->ralloc= malloc(bf->s_ralloc*sizeof(bf->ralloc[0]));\n"
"  }\n"
"  R= bf->ralloc[bf->n_ralloc++]= malloc(size);\n"
"  return R;\n"
"}\n"
"\n"
"static void *bff_ralloc0(bff_t *bf, uint32_t size)\n"
"{\n"
"  void *R;\n"
"  R= bff_ralloc(bf,size);\n"
"  memset(R, 0, size);\n"
"  return R;\n"
"}\n"
"\n"
"static void bff_free_ralloc(bff_t *bf)\n"
"{\n"
"  int i;\n"
"  for(i=0;i<bf->n_ralloc;i++)\n"
"   free(bf->ralloc[i]);\n"
"  bf->n_ralloc= 0;\n"
"}\n"
"\n"
"static void bff_reset_ralloc(bff_t *bf)\n"
"{\n"
"  bf->n_ralloc= 0;\n"
"}\n"
"\n"
"\n"
"\n"
"static void bff_bwrite_u8(bff_t *bf, uint8_t V)\n"
"{\n"
"  if (bf->w_pass)\n"
"  {\n"
"    bf->o_pos[0]= V;\n"
"  }\n"
"  bf->o_pos++;\n"
"}\n"
"\n"
"static void bff_bwrite_u16(bff_t *bf, uint16_t V)\n"
"{\n"
"  if (bf->w_pass)\n"
"  {\n"
"    bf->o_pos[0]= V;\n"
"    bf->o_pos[1]= V>>8;\n"
"  }\n"
"  bf->o_pos+= 2;\n"
"}\n"
"\n"
"static void bff_bwrite_u32(bff_t *bf, uint32_t V)\n"
"{\n"
"  if (bf->w_pass)\n"
"  {\n"
"    bf->o_pos[0]= V;\n"
"    bf->o_pos[1]= V>>8;\n"
"    bf->o_pos[2]= V>>16;\n"
"    bf->o_pos[3]= V>>24;\n"
"  }\n"
"  bf->o_pos+= 4;\n"
"}\n"
"\n"
"static void bff_bwrite_u64(bff_t *bf, uint64_t V)\n"
"{\n"
"  int i;\n"
"  if (bf->w_pass)\n"
"  {\n"
"    for(i=0;i<8;i++)\n"
"    {\n"
"      bf->o_pos[i]= V;\n"
"      V>>= 8;\n"
"    }\n"
"  }\n"
"  bf->o_pos+= 8;\n"
"}\n"
"\n"
"static void bff_bwrite_i8(bff_t *bf, int8_t V)\n"
"   { bff_bwrite_u8(bf,(uint8_t)V); }\n"
"\n"
"static void bff_bwrite_i16(bff_t *bf, int16_t V)\n"
"   { bff_bwrite_u16(bf,(uint16_t)V); }\n"
"\n"
"static void bff_bwrite_i32(bff_t *bf, int32_t V)\n"
"   { bff_bwrite_u32(bf,(uint32_t)V); }\n"
"\n"
"static void bff_bwrite_i64(bff_t *bf, int64_t V)\n"
"   { bff_bwrite_u64(bf,(uint64_t)V); }\n"
"\n"
"\n"
"static void bff_bwrite_s8(bff_t *bf, uint8_t *D)\n"
"{\n"
"  uint32_t i,L;\n"
"  L= D ? 0 : strlen((char*) D);\n"
"  bff_bwrite_u32(bf, L);\n"
"  if (bf->w_pass)\n"
"  {\n"
"    for(i=0;i<L;i++) bf->o_pos[i]= D[i];\n"
"  }\n"
"  bf->o_pos+= L;\n"
"}\n"
"\n"
"static void bff_bwrite_s16(bff_t *bf, uint16_t *D)\n"
"{\n"
"  uint32_t i,L;\n"
"  L= D ? 0 : str16len(D);\n"
"  bff_bwrite_u32(bf, L);\n"
"  if (bf->w_pass)\n"
"  {\n"
"    for(i=0;i<L;i++) \n"
"    {\n"
"      bf->o_pos[2*i+0]= D[i]; \n"
"      bf->o_pos[2*i+1]= D[i]>>8; \n"
"    }\n"
"  }\n"
"  bf->o_pos+= 2*L;\n"
"}\n"
"\n"
"static void bff_bwrite_s32(bff_t *bf, uint32_t *D)\n"
"{\n"
"  uint32_t i,L;\n"
"  L= D ? 0 : str32len(D);\n"
"  bff_bwrite_u32(bf, L);\n"
"  if (bf->w_pass)\n"
"  {\n"
"    for(i=0;i<L;i++) \n"
"    {\n"
"      bf->o_pos[4*i+0]= D[i]; \n"
"      bf->o_pos[4*i+1]= D[i]>>8; \n"
"      bf->o_pos[4*i+2]= D[i]>>16; \n"
"      bf->o_pos[4*i+3]= D[i]>>24; \n"
"    }\n"
"  }\n"
"  bf->o_pos+= 4*L;\n"
"}\n"
"\n"
"static int str16len(uint16_t *S)\n"
"{\n"
"  int i;\n"
"  for(i=0;S[i];i++) ;\n"
"  return i;\n"
"}\n"
"\n"
"static int str32len(uint32_t *S)\n"
"{\n"
"  int i;\n"
"  for(i=0;S[i];i++) ;\n"
"  return i;\n"
"}\n"
"\n"
"\n"
"static void bff_alloc_o_buffer(bff_t *bf, uint64_t size)\n"
"{\n"
"  uint64_t delta;\n"
"  if (bf->o_size>=size) return ;\n"
"\n"
"  delta= size - bf->o_size;\n"
"  if (delta*10<bf->o_size)\n"
"    delta+= bf->o_size/2;\n"
"  bf->o_size+= delta;\n"
"\n"
"  if (bf->o_buffer)\n"
"     bf->o_buffer= realloc(bf->o_buffer, bf->o_size);\n"
"  else\n"
"     bf->o_buffer= malloc(bf->o_size);\n"
"}\n"
"\n"
"static int bff_encode_and_write\n"
"   (bff_t *bf, void *obj, uint32_t otype,\n"
"    void (*encode)(bff_t *,void*), uint64_t *objidptr)\n"
"{\n"
"  uint8_t K[1];\n"
"  uint64_t osize;\n"
"  bf->w_pass= 0;\n"
"  bf->o_pos= K;\n"
"  (*encode)(bf, obj);\n"
"  osize= bf->o_pos - K;\n"
"  bff_alloc_o_buffer(bf, osize);\n"
"  bf->o_pos= bf->o_buffer;\n"
"  bf->w_pass= 1;\n"
"  (*encode)(bf, obj);\n"
"\n"
"  if (*objidptr)\n"
"  {\n"
"    bff_free_object(bf, *objidptr);\n"
"  }\n"
"  else\n"
"  {\n"
"    *objidptr= bf->next_avail_ident++;\n"
"  }\n"
"  return bff_write_object(bf, *objidptr, otype, osize, bf->o_buffer);\n"
"}\n"
"\n"
"static int bff_continue_postponed_writes(bff_t *bf)\n"
"{\n"
"  postponed_write_t pw;\n"
"  if (postpone_buffer_read(bf->pb, &pw, 1)!=1) return 0;\n"
"  return (*pw.func)(bf, pw.obj);\n"
"}\n"
"\n"
"\n"
"static void bff_pointercheck\n"
"   (bff_t *bf, void *obj, uint64_t *objid, int (*writefunc)(void *, void*))\n"
"{\n"
"  postponed_write_t pw;\n"
"  if (*objid) return ;\n"
"  *objid= bf->next_avail_ident++;\n"
"  pw.obj= obj;\n"
"  pw.func= writefunc;\n"
"  postpone_buffer_put(bf->pb, pw);\n"
"}\n"
"\n"
"\n"
"static void bff_init_floio(bff_t *bf)\n"
"{\n"
"  uint64_t T;\n"
"  T= (uint64_t) bf->floio_store;\n"
"  T%= 16;\n"
"  if (T)  bf->floio_cptr= bf->floio_store + (16-T);\n"
"    else  bf->floio_cptr= bf->floio_store;\n"
"\n"
"  bf->floio_fptr= (float*) bf->floio_cptr;\n"
"  bf->floio_dptr= (double*) bf->floio_cptr;\n"
"\n"
"  flolay_get(bf->floio_perm_fw, bf->floio_perm_fr,\n"
"             bf->floio_perm_dw, bf->floio_perm_dr);\n"
"}\n"
"\n"
"static int bff_bread_f32(bff_t *bf, float *R)\n"
"{\n"
"  int i;\n"
"  if (bf->o_pos+4>bf->o_end) bff_fail_bread(bf);\n"
"  for(i=0;i<4;i++)\n"
"    bf->floio_cptr[(int) bf->floio_perm_fr[i]]= bf->o_pos[i];\n"
"  *R= *bf->floio_fptr;\n"
"  bf->o_pos+= 4;\n"
"  return 0;\n"
"}\n"
"\n"
"static void bff_bwrite_f32(bff_t *bf, float V)\n"
"{\n"
"  if (bf->w_pass)\n"
"  {\n"
"    int i;\n"
"    *bf->floio_fptr= V;\n"
"    for(i=0;i<4;i++)\n"
"      bf->o_pos[(int) bf->floio_perm_fw[i]]= bf->floio_cptr[i];\n"
"  }\n"
"  bf->o_pos+= 4;\n"
"}\n"
"\n"
"\n"
"static int bff_bread_f64(bff_t *bf, double *R)\n"
"{\n"
"  int i;\n"
"  if (bf->o_pos+8>bf->o_end) bff_fail_bread(bf);\n"
"  for(i=0;i<8;i++)\n"
"    bf->floio_cptr[(int) bf->floio_perm_dr[i]]= bf->o_pos[i];\n"
"  *R= *bf->floio_dptr;\n"
"  bf->o_pos+= 8;\n"
"  return 0;\n"
"}\n"
"\n"
"static void bff_bwrite_f64(bff_t *bf, double V)\n"
"{\n"
"  if (bf->w_pass)\n"
"  {\n"
"    int i;\n"
"    *bf->floio_dptr= V;\n"
"    for(i=0;i<8;i++)\n"
"      bf->o_pos[(int) bf->floio_perm_dw[i]]= bf->floio_cptr[i];\n"
"  }\n"
"  bf->o_pos+= 8;\n"
"}\n"
"\n"

#line 362 "/home/dodo/tmp/BB/bff-1.1/src/gen.u"
  );
}

#line 371 "/home/dodo/tmp/BB/bff-1.1/src/gen.u"

static void gen_write_function
     (FILE *out, defile_t *fil, odef_t *od, int body)
{
  fprintf(out,"int %swrite_%s(bff_t *bf,%s%s_t *obj)", 
       fil->pfx, od->name, fil->pfx, od->name);
  if (!body)
  {
    fprintf(out,";\n");
    return ;
  }
  fprintf(out,"\n{\n");
  fprintf(out, 
  "  if (bff_encode_and_write(bf,obj,%d, (void*) %sencode_%s, &obj->_objid))\n"
  "    return 1;\n",
    od->typenumber, fil->pfx, od->name);
  fprintf(out, 
  "  return bff_continue_postponed_writes(bf);\n");
  fprintf(out,"}\n");
}

static void enumerate_objects(defile_t *fil)
{
  int last;
  odef_t *od;
  last= 32;
  for(od=fil->odefs.next;od!=&fil->odefs;od=od->next)
    od->typenumber= last++;
}

static void gen_encode_function
    (FILE *out, defile_t *fil, odef_t *od,int body)
{
  field_t *f;
  fprintf(out,"static void %sencode_%s(bff_t *bf, %s%s_t *obj)",
            fil->pfx, od->name, fil->pfx, od->name);
  if (!body)
  {
    fprintf(out, ";\n");
    return ;
  }
  fprintf(out,"\n{\n");
  gen_array_index_declarations(out, od);
  for(f=od->fields.next;f!=&od->fields;f=f->next)
    gen_field_writer(out, fil, od, f);
  fprintf(out,"}\n");
}

static void gen_field_writer
    (FILE *out, defile_t *fil, odef_t *od, field_t *f)
{
  char buf[2000];
  type_t *T;
  odef_t *S;
  T= f->type;
  switch(T->kind)
  {
  case TYPE_ARRAY:
    gen_array_loop(out, T);
    sprintf(buf,"&obj->%s", f->name); 
    gen_array_reference(out, buf, T);
    gen_elementary_writer(out, fil, T->ref, buf);
    return ;
  case TYPE_DARRAY:
    fprintf(out,"  bff_bwrite_u32(bf, obj->N%s);\n", f->name);
    fprintf(out,"  for(i0=0;i0<obj->N%s;i0++)\n", f->name);
    sprintf(buf, "&obj->%s[i0]", f->name);
    gen_elementary_writer(out, fil, T->ref, buf);
    return ;
  case TYPE_POINTER:
    S= T->ref->obdef;
    if (T->ref->kind!=TYPE_SUB || !S->top_level)
      exit(fprintf(stderr,"error while generating write code for field %s\n"
                          "in object %s: referred data is not an object\n",
           f->name, od->name));
    fprintf(out,
    "  if (&obj->%s)\n"
    "    bff_pointercheck(bf, obj->%s, &obj->%s->_objid, (void*) %swrite_%s);\n",
       f->name, f->name, f->name, fil->pfx, S->name);
    fprintf(out,
    "  bff_bwrite_u64(bf, obj->ID%s);\n", f->name);
    return ;
  }
  sprintf(buf, "&obj->%s", f->name);
  gen_elementary_writer(out, fil, T, buf);
}

static void gen_elementary_writer
  (FILE *out, defile_t *fil, type_t *T, char *ref)
{
  char buf[200];
  if (type_to_str(T, buf)) 
  {
    odef_t *O;
    if (T->kind!=TYPE_SUB) return ;
    O= T->obdef;
    fprintf(out,"  %sencode_%s(bf,%s);\n", fil->pfx,O->name, ref);
    return ;
  }
  // SUBTLE !! SUBTLE !! SUBTLE !! SUBTLE !! SUBTLE !! SUBTLE !! SUBTL
  // The string 'ref' contains an expression of pointer type,
  // which uses the addressof operator on a field expression.
  // the following +1 removes that operator and passes the field
  // by value.
  fprintf(out,"  bff_bwrite_%s(bf,%s);\n", buf, ref+1);
  //
  // SUBTLE !! SUBTLE !! SUBTLE !! SUBTLE !! SUBTLE !! SUBTLE !! SUBTL
}

#line 488 "/home/dodo/tmp/BB/bff-1.1/src/gen.u"

static void gen_magic(FILE *out, defile_t *fil)
{
  int i;
  fprintf(out,"static const uint8_t %smagic[16]= {\n", fil->pfx);
  for(i=0;i<16;i++) fprintf(out,"0x%02x, ", fil->magic[i]);
  fprintf(out,"\n};\n");
}

static int proto_or_body(FILE *out, int body)
{
  if (body) fprintf(out,"\n{\n");
       else fprintf(out,";\n");
  return body;
}

static int end_body(FILE *out)
{
  fprintf(out,"}\n");
}

static void gen_util_functions(FILE *out, defile_t *fil,int body)
{
  fprintf(out,"int %sremove_object(bff_t *bff, uint64_t objid)", fil->pfx);
  if (proto_or_body(out,body))
  {
    fprintf(out, "  return bff_free_object(bff, objid);\n");
    end_body(out);
  }
  fprintf(out,"int %sopen(const char *fn, int mode,bff_t **R)", fil->pfx);
  if (proto_or_body(out,body))
  {
    fprintf(out, 
    "  return _bff_open(fn, mode, R, %smagic);\n", fil->pfx);
    end_body(out);
  }

  fprintf(out,"int %screate(const char *fn, bff_t **R)", fil->pfx);
  if (proto_or_body(out,body))
  {
    fprintf(out, 
    "  return _bff_create(fn, %smagic, R);\n", fil->pfx);
    end_body(out);
  }

  fprintf(out,"void %sclose(bff_t *bf)", fil->pfx);
  if (proto_or_body(out,body))
  {
    fprintf(out, "  _bff_close(bf);\n");
    end_body(out);
  }

  fprintf(out,"int %serrno(bff_t *bf)", fil->pfx);
  if (proto_or_body(out,body))
  {
    fprintf(out, "  return bf->eno;\n");
    end_body(out);
  }

}

#line 150 "/home/dodo/tmp/BB/bff-1.1/src/main.u"

void dbg_print_file(defile_t *df)
{
  dbg_print_objects(&df->odefs);
}

static void dbg_print_objects(odef_t *all)
{
  odef_t *d;
  field_t *f;
  raw_t *r;
  for(d=all->next;d!=all;d=d->next)
  {
    printf("object(%s)\n", d->name);
    for(f=d->fields.next;f!=&d->fields;f=f->next)
    {
      printf("   field(%s): ", f->name);
      dbg_print_type(f->type);
      printf("\n");
    }
    for(r=d->raws.next;r!=&d->raws;r=r->next)
      printf("   raw %s\n", r->text);
  }
}

static void dbg_print_type(type_t *t)
{
  int i;
  odef_t *O;
  switch(t->kind)
  {
   case TYPE_I8: printf("int8"); return ;
   case TYPE_U8: printf("unt8"); return ;
   case TYPE_I16: printf("int16"); return ;
   case TYPE_U16: printf("unt16"); return ;
   case TYPE_I32: printf("int32"); return ;
   case TYPE_U32: printf("unt32"); return ;
   case TYPE_I64: printf("int64"); return ;
   case TYPE_U64: printf("unt64"); return ;
   case TYPE_F32: printf("flo32"); return ;
   case TYPE_F64: printf("flo64"); return ;
   case TYPE_S8:  printf("str8"); return ;
   case TYPE_S16:  printf("str16"); return ;
   case TYPE_S32:  printf("str32"); return ;
   case TYPE_DARRAY: printf("darray "); break;
   case TYPE_POINTER: printf("ptr "); break;
   case TYPE_ARRAY: printf("array(");
          for(i=0;i<t->n_dimensions;i++)
            if (i) printf(",%d", t->dimensions[i]);
             else printf("%d",t->dimensions[i]);
          printf(") "); break;
   case TYPE_UNRESOLVED: printf("unresolved(%s)", t->name); return ;
   case TYPE_SUB: O= t->obdef; printf("%s", O->name); return ;
   default: printf("unknown(%d)", t->kind); return ;
  }
  if (t->ref) dbg_print_type(t->ref);
}

#line 134 "/home/dodo/tmp/BB/bff-1.1/src/main.u"

int main(int argc, char **argv)
{
  defile_t *df;
  FILE *h_out, *c_out;
  df= parse_inputs(argc,argv);
  // dbg_print_file(df);
  if (!df->pfx)   df->pfx= "bff_";
  if (!df->hname) df->hname= "bfo.h";
  if (!df->sname) df->sname= "bfo.c";
  enumerate_objects(df);
  gen_all(df);
  return 0;
}

