#line 355 "/home/dodo/tmp/BB/bff-1.1/src/parse.u"
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <errno.h>
#include "unquote.h"
#include "parse.h"

#line 9 "parse.c"
static int get_line(FILE *f,int siz,char *buf, int *lineno);
static void make_raw(raw_t *dst, char *txt,char *fn,int lineno);
static int collect_lines
   (raw_t *dst, char *fn, FILE *f, int lineno, char *start,char *end);
static int get_word(char *x, char **R, char **rest);
static int get_quot(char *x,char **R,char **rest);
static void syntax_error
   (const char *fn, char *line, int lineno, const char *msg);
static char *make_src(char *fn,int lineno);
static int is_space(char *x);
static int is_letter(char *x);
static int is_letter_or_digit(char *x);
static type_t *parse_type(char *x);
static char *collect_par(char *x,char **rest);
static char *collect_str(char *a,char *b);
static int parse_dimensions(char *x, int *Rn, int **Rd);
static void resolve_unknown_types(odef_t *all, odef_t *OD);
static int keycmp(char *start, char *end, char *key);
#line 4 "/home/dodo/tmp/BB/bff-1.1/src/parse.u"
static int get_line(FILE *f,int siz,char *buf, int *lineno)
{
  int i;
again:
  buf[siz-1]= 0;
  if (!fgets(buf,siz-1,f)) return 1;
  (*lineno)++;
  for(i=0;i<siz;i++) 
    if (buf[i]=='\n' || buf[i]=='\r' || buf[i]=='#')
     { buf[i]=0; break; }
  if (!buf[0]) goto again;
  for(i=0;buf[i];i++)
    if (buf[i]>' ' && buf[i]<127) return 0;
  goto again;
}

static void make_raw(raw_t *dst, char *txt,char *fn,int lineno)
{
  raw_t *R;
  R= calloc(1,sizeof(*R));
  R->src= make_src(fn,lineno);
  R->text= txt;
  R->next= dst;
  R->prev= dst->prev;
  dst->prev->next= R;
  dst->prev= R;
}

static int collect_lines
   (raw_t *dst, char *fn, FILE *f, int lineno, char *start,char *end)
{
  int o_line;
  char buf[2000];
  int i,j;
  o_line= lineno;
  while(1)
  {
    if (!fgets(buf,sizeof(buf),f)) 
     exit(fprintf(stderr,"%s:%d: unterminated %s\n", o_line, start));
    lineno++;
    for(i=0;buf[i] && end[i];i++) 
      if (buf[i]!=end[i]) break;
    for(j=i;buf[j];j++)
      if (buf[j]>' ' && buf[j]<127) break;
    if (end[i] || buf[j])
    {
      make_raw(dst, strdup(buf), fn, lineno);
      continue;
    }
    break;
  }
  return lineno;
}

#line 297 "/home/dodo/tmp/BB/bff-1.1/src/parse.u"
static int get_word(char *x, char **R, char **rest)
{
  char *z,*K;
  int L;
  while(is_space(x)) x++;
  if (!x[0]) return 1;
  z= x;
  while(!is_space(z)) z++;
  L= z - x;
  *rest= z;
  K= malloc(L+1);
  memcpy(K,x,L);
  K[L]= 0;
  *R= K;
  return 0;
}

static int get_quot(char *x,char **R,char **rest)
{
  char *z;
  while(is_space(x)) x++;
  if (x[0]!='"') return 1;
  x++;
  for(z=x;z[0];z++)
   if (z[0]=='\\') z++;
   else if (z[0]=='"') break;
  if (!z[0]) return 1;
  *rest= z+1;
  *R= unquote_c_literal_iloz(x-1, z-x+1);
  return 0;
}

static void syntax_error
   (const char *fn, char *line, int lineno, const char *msg)
{
  fprintf(stderr,"%s:%d: %s\n", fn, lineno,msg);
  fprintf(stderr,"%s\n",line);
  exit(1);
}

static char *make_src(char *fn,int lineno)
{
  char buf[2000];
  snprintf(buf,sizeof(buf),"%s:%d",fn,lineno);
  buf[sizeof(buf)-1]= 0;
  return strdup(buf);
}

#line 383 "/home/dodo/tmp/BB/bff-1.1/src/parse.u"

static int is_space(char *x)
{
  if (*x==0) return 0;
  else if (*x<=' ' || *x>=127) return 1;
  else return 0;
}

static int is_letter(char *x)
{
  if (*x>='a' && *x<='z') return 1;
  if (*x>='A' && *x<='Z') return 1;
  if (*x=='_') return 1;
  return 0;
}

static int is_letter_or_digit(char *x)
{
  if (*x>='0' && *x<='9') return 1;
  return is_letter(x);
}

static type_t *parse_type(char *x)
{
  type_t *T,**ref;
  char *z;

  while(is_space(x)) x++;
  if (!x[0]) return NULL;
  z= x;
  if (!is_letter(z)) return NULL;
  while(is_letter_or_digit(z)) z++;
  if (z==x) return NULL;

  T= calloc(1,sizeof(*T));
  ref= NULL;

  if (!keycmp(x,z,"array"))
  {
    char *par;
    par= collect_par(z, &z);
    if (!par) return NULL;
    if (!strcmp(par,"*")) 
      T->kind= TYPE_DARRAY;
    else
    {
      T->kind= TYPE_ARRAY;
      if (parse_dimensions(par,&T->n_dimensions,&T->dimensions))
        return NULL;
    }
    ref= &T->ref;
  }
  else if (!keycmp(x,z,"ptr"))
  {
    T->kind= TYPE_POINTER;
    ref= &T->ref;
  }
#define BASIC(a,b) else if (!keycmp(x,z,a)) T->kind=b; 
  BASIC("int16", TYPE_I16)
  BASIC("unt16", TYPE_U16)
  BASIC("int8", TYPE_I8)
  BASIC("unt8", TYPE_U8)
  BASIC("int32", TYPE_I32)
  BASIC("unt32", TYPE_U32)
  BASIC("int64", TYPE_I64)
  BASIC("unt64", TYPE_U64)
  BASIC("flo32", TYPE_F32)
  BASIC("flo64", TYPE_F64)
  BASIC("str8", TYPE_S8)
  BASIC("str16", TYPE_S16)
  BASIC("str32", TYPE_S32)
#undef BASIC
  else { T->kind= TYPE_UNRESOLVED; T->name= collect_str(x,z); }

  if (ref)
  {
    *ref= parse_type(z);
    if (!*ref) return NULL;
    switch(T->kind)
    {
    case TYPE_POINTER:
      if (T->ref->kind!=TYPE_UNRESOLVED) return NULL;
      break;
    case TYPE_ARRAY:
    case TYPE_DARRAY:
      if (T->ref->kind==TYPE_ARRAY || T->ref->kind==TYPE_DARRAY
          || T->ref->kind==TYPE_POINTER)
       return NULL;
      break;
    }
    return T;
  }

  while(is_space(z)) z++; 
  if (z[0]) return NULL; // extra chars at the end
  return T;
}

static char *collect_par(char *x,char **rest)
{
  char *z,*R,*p;
  while(is_space(x)) x++;
  if (x[0]!='(') return NULL;
  z= x+1;
  for(z=x+1;z[0];z++)
    if (z[0]==')') break;
  if (!z[0]) return NULL;
  *rest= z+1;
  p= R= malloc(z-x+1);
  for(x=x+1;x!=z;x++)
   if (!is_space(x)) *(p++)= *x;
  *p= 0;
  return R;
}

static char *collect_str(char *a,char *b)
{
  char *R;
  int L;
  L= b-a;
  R= malloc(L+1);
  memcpy(R, a, L);
  R[L]= 0;
  return R;
}

static int parse_dimensions(char *x, int *Rn, int **Rd)
{
  int *K;
  int n,i,z;
  z= 0;
  for(i=0;x[i];i++)
  {
    if (x[i]==',')
    {
     if (i==0 || x[i+1]==',' || x[i+1]==0) return 1;
     z++;
     continue;
    }
    if (x[i]>='0' && x[i]<='9') continue; 
    return 1; 
  }
  K= calloc(z+1,sizeof(int));
  n= 0;
  for(i=0;x[i];i++)
  {
    if (x[i]==',') { n++; continue; }
    K[n]= K[n]*10 + x[i]-'0';
  }
  *Rn= n+1;
  *Rd= K;
  return 0;
}


static void resolve_unknown_types(odef_t *all, odef_t *OD)
{
  field_t *F;
  odef_t *S;
  type_t *T;
  int no_recurse, only_top;

  if (OD->aux&1)
  {
    fprintf(stderr,"a loop was detected regarding the definitions of the "
                   "following objects:\n");
    for(S=all->next;S!=all;S=S->next)
       if (S->aux&1) printf(" %s", S->name);
    printf("\n");
    exit(1);
  }
  if (OD->aux&2) return ;
  OD->aux= 3;
  for(F=OD->fields.next;F!=&OD->fields;F=F->next)
  {
    T= F->type;
    no_recurse= 0;
    only_top= 0;
again:
    switch(T->kind)
    {
    case TYPE_ARRAY:
    case TYPE_DARRAY:
        T= T->ref;
        goto again;
    case TYPE_POINTER:
        T= T->ref;
        no_recurse= 1;
        only_top= 1;
        goto again;
    case TYPE_UNRESOLVED:
       for(S=all->next;S!=all;S=S->next)
         if (!strcmp(S->name,T->name)) break;
       if (S==all)
          exit(fprintf(stderr,"in definition of %s: unknown type %s\n",
                OD->name, T->name));
       if (only_top && !S->top_level)
          exit(fprintf(stderr,"in definition of %s: only "
                  "pointers to objects are allowed\n",
                OD->name));
       T->kind= TYPE_SUB;
       T->obdef= S;
       if (!no_recurse) 
          resolve_unknown_types(all, S);
       break;
    }
  }
  OD->aux= 2;
}

static int keycmp(char *start, char *end, char *key)
{
  while(start!=end) if (*start++!=*key++) return 1;
  return 0;
}

#line 59 "/home/dodo/tmp/BB/bff-1.1/src/parse.u"

defile_t* parse_file(char* fn)
{
  FILE *f; 
  odef_t *OD;
  defile_t *DF;
  char g[2000];
  int lineno;
  lineno= 0;
  f= fopen(fn,"rb");
  if (!f)
    exit(fprintf(stderr,"can't open %s: %s\n", fn, strerror(errno)));
  DF= calloc(1,sizeof(*DF));

#define llinit(X) X.prev= X.next= &X
  llinit(DF->src_top);
  llinit(DF->src_bot);
  llinit(DF->hdr_top);
  llinit(DF->hdr_bot);
  llinit(DF->odefs);

  OD= NULL;
  while(!get_line(f,sizeof(g),g,&lineno))
  {
    char *k,*z;
    if (get_word(g,&k,&z))
     syntax_error(fn,g,lineno,"missing command");
#line 165 "/home/dodo/tmp/BB/bff-1.1/src/parse.u"
    if (!strcmp(k,"output_source"))
    {
      if (get_quot(z,&k,&z)) syntax_error(fn,g,lineno,"file name missing"); 
      DF->sname= k;
      continue;
    }
    if (!strcmp(k,"output_header"))
    {
      if (get_quot(z,&k,&z)) syntax_error(fn,g,lineno,"file name missing"); 
      DF->hname= k;
      continue;
    }
    if (!strcmp(k,"src_top"))
    {
      lineno= collect_lines(&DF->src_top, fn,
                            f, lineno, "src_top", "end_src_top");
      continue;
    }
    if (!strcmp(k,"src_bot"))
    {
      lineno= collect_lines(&DF->src_bot, fn,
                            f, lineno, "src_bot", "end_src_bot");
      continue;
    }
    if (!strcmp(k,"hdr_bot"))
    {
      lineno= collect_lines(&DF->hdr_bot, fn,
                            f, lineno, "hdr_bot", "end_hdr_bot");
      continue;
    }
    if (!strcmp(k,"hdr_top"))
    {
      lineno= collect_lines(&DF->hdr_top, fn,
                            f, lineno, "hdr_top", "end_hdr_top");
      continue;
    }
    if (!strcmp(k,"library_prefix"))
    {
      if (get_word(z,&k,&z)) syntax_error(fn,g,lineno,"prefix missing");
      DF->pfx= k;
      continue;
    }
    if (!strcmp(k,"magic"))
    {
      int L;
      if (get_quot(z,&k,&z)) syntax_error(fn,g,lineno,"magic missing");
      L= strlen(k);
      if (L>MAGICLEN) syntax_error(fn,g,lineno,"magic too long");
      memset(DF->magic,0,MAGICLEN);
      memcpy(DF->magic,k,L);
      DF->has_magic= 1;
      continue;
    }
    if (!strcmp(k,"version"))
    {
      int M,L,D,i,major,minor;
      if (get_word(z,&k,&z)) syntax_error(fn,g,lineno,"version missing");
      L= strlen(k);
      M= D= 0;
      for(i=0;k[i];i++)
        if (k[i]=='.') M++;
        else if (isdigit(k[i])) D++;
      if (M!=1 || (D+M!=L) || k[0]=='.' || k[L-1]=='.')
        syntax_error(fn,g,lineno,"bad version number");
      major= atoi(k);
      for(i=0;k[i];i++)
        if (k[i]=='.') { minor= atoi(k+i+1); break; }
      DF->version= (major<<16) | minor;
      continue;
    }

#line 237 "/home/dodo/tmp/BB/bff-1.1/src/parse.u"
    if (!strcmp(k,"object") || !strcmp(k,"struct"))
    {
      odef_t *d;
      int top_level;
      top_level= !strcmp(k,"object");
      if (OD) syntax_error(fn,g,lineno,"unterminated object definition");
      if (get_word(z,&k,&z)) syntax_error(fn,g,lineno,"object name missing");
      for(d=DF->odefs.next;d!=&DF->odefs;d=d->next)
        if (!strcmp(d->name,k))
        {
          fprintf(stderr,"%s: previous declaration was here\n", d->src);
          syntax_error(fn,g,lineno,"redeclaration of object\n");
        }
      OD= calloc(1,sizeof(*OD));
      OD->top_level= top_level;
      OD->next= &DF->odefs;
      OD->prev= DF->odefs.prev;
      DF->odefs.prev->next= OD;
      DF->odefs.prev= OD;
      OD->src= make_src(fn,lineno);
      OD->name= k;
      OD->raws.prev= OD->raws.next= &OD->raws;
      OD->fields.prev= OD->fields.next= &OD->fields;
      continue;
    }

#line 264 "/home/dodo/tmp/BB/bff-1.1/src/parse.u"
  if (!strcmp(k,"field"))
  {
    field_t *d;
    if (get_word(z,&k,&z)) syntax_error(fn,g,lineno,"field name missing");
    if (!OD) syntax_error(fn,g,lineno,"dangling 'field' command");
    for(d=OD->fields.next;d!=&OD->fields;d=d->next)
     if (!strcmp(d->name,k))
     {
       fprintf(stderr,"%s: previous declaration was here\n",d->src);
       syntax_error(fn,g,lineno, "multiple declaration of the same field\n");
     }
    d= calloc(1,sizeof(*d));
    d->name= k;
    d->prev= OD->fields.prev;
    d->next= &OD->fields;
    OD->fields.prev->next= d;
    OD->fields.prev= d;
    d->type= parse_type(z);
    if (!d->type)
      syntax_error(fn,g,lineno,"bad type");
    d->src= make_src(fn,lineno);
    continue;
  }

#line 289 "/home/dodo/tmp/BB/bff-1.1/src/parse.u"
  if (!strcmp(k,"raw"))
  {
    if (!OD) syntax_error(fn,g,lineno,"dangling 'raw' command");
    make_raw(&OD->raws, strdup(z), fn, lineno);
    continue;
  }

#line 90 "/home/dodo/tmp/BB/bff-1.1/src/parse.u"
    if (!strcmp(k,"end"))
    {
      if (!OD) syntax_error(fn,g,lineno,"dangling 'end' command");
      OD= NULL;
      continue;
    }
    syntax_error(fn,g,lineno,"unknown command");
  }
  if (OD)
    exit(fprintf(stderr,"%s: unterminated object definition\n",OD->src));
  fclose(f);
  return DF;
}


#line 600 "/home/dodo/tmp/BB/bff-1.1/src/parse.u"

void resolve_unknown(defile_t *d)
{
  odef_t *od;
  for(od=d->odefs.next;od!=&d->odefs;od=od->next)
     resolve_unknown_types(&d->odefs, od);
}

int type_to_str(type_t *T,char *o)
{
  switch(T->kind)
  {
    case  TYPE_I8:  strcpy(o,"i8");  return 0;
    case  TYPE_U8:  strcpy(o,"u8");  return 0;
    case  TYPE_I16: strcpy(o,"i16"); return 0;
    case  TYPE_U16: strcpy(o,"u16"); return 0;
    case  TYPE_I32: strcpy(o,"i32"); return 0;
    case  TYPE_U32: strcpy(o,"u32"); return 0;
    case  TYPE_I64: strcpy(o,"i64"); return 0;
    case  TYPE_U64: strcpy(o,"u64"); return 0;
    case  TYPE_F32: strcpy(o,"f32"); return 0;
    case  TYPE_F64: strcpy(o,"f64"); return 0;
    case  TYPE_S8:  strcpy(o,"s8");  return 0;
    case  TYPE_S16: strcpy(o,"s16"); return 0;
    case  TYPE_S32: strcpy(o,"s32"); return 0;
  }
  return 1;
}

