#line 2 "of.test.u"
#include <stdlib.h>
#include <stdio.h>
#include <stdint.h>
#include <stdarg.h>
#include <string.h>
#include <time.h>
#include <fcntl.h>
#include <unistd.h>
#include <errno.h>
#include <setjmp.h>


typedef struct _bff bff_t;

#line 402 "of.u"

#define BLOCK_HEADER_SIZE     24
#define OBJECT_HEADER_SIZE     8
#define FREE_BLOCK_THRESHOLD  (2*(BLOCK_HEADER_SIZE+OBJECT_HEADER_SIZE))
#define FILE_HEADER_SIZE      32

#line 362 "of.u"

typedef struct { uint64_t offset, size; int qindex; } free_block_t;
typedef struct { uint64_t ident, type, offset; } object_descriptor_t;

#line 28 "of.test.c"
typedef struct fbm_tree fbm_t;


static fbm_t *fbm_new    ();
static free_block_t  *fbm_find   (fbm_t *T, uint64_t key);
static free_block_t  *fbm_insert (fbm_t *T, free_block_t *data);
static free_block_t  *fbm_remove (fbm_t *T, uint64_t key);
static free_block_t  *fbm_destroy(fbm_t *T);



typedef struct fbm_node {
   struct fbm_node *left, *right,*parent;
   int height;
   free_block_t *data;
} fbm_node_t;

struct fbm_tree { fbm_node_t *root; };


static int fbm_compare(uint64_t A, uint64_t B) { return (A<B?-1:(A>B?1:0)); }
static uint64_t fbm_upper(free_block_t *A)        { return (A->offset+A->size-1); }
static uint64_t fbm_lower(free_block_t *A)        { return A->offset; }
static int fbm_HEIGHT(fbm_node_t * A) { return A? A->height : 0; }

static void fbm_calculate_height(fbm_node_t *N)
{
  int a,b;
  a= fbm_HEIGHT(N->left);
  b= fbm_HEIGHT(N->right);
  N->height= (a>b?a:b)+1;
}

static void fbm_replace_child_in_parent 
      (fbm_t* T, fbm_node_t* old, fbm_node_t* nw)
{
  if (old->parent)
    {
      if (old->parent->left == old)
	old->parent->left = nw;
      else
	old->parent->right = nw;
    }
  else
    {
      T->root = nw;
    }
  if (nw) nw->parent= old->parent;
}

static void fbm_set_left_child(fbm_node_t *P, fbm_node_t *N)
{
  P->left= N; 
  if (N) N->parent= P;
}

static void fbm_set_right_child(fbm_node_t *P, fbm_node_t *N)
{
  P->right= N; 
  if (N) N->parent= P;
}

static fbm_node_t *fbm_find_node(fbm_t *T, uint64_t key)
{
   fbm_node_t *n;
   n= T->root;
   while (n)
   {
     int u;
     u= fbm_compare(key, fbm_lower(n->data));
     if (u==0) return n;
     if (u<0)  { n= n->left; continue; }
     u= fbm_compare(key, fbm_upper(n->data));
     if (u<=0) return n;
     n= n->right;
   }
   return NULL;
}

static int fbm_right_heavy(fbm_node_t *A)
{
  if (!A) return 0;
  return (fbm_HEIGHT(A->right)-fbm_HEIGHT(A->left)) > 0;
}

static int fbm_left_heavy(fbm_node_t *A)
{
  if (!A) return 0;
  return (fbm_HEIGHT(A->right)-fbm_HEIGHT(A->left)) < 0;
}

static void fbm_rotate_left(fbm_t *T, fbm_node_t *A)
{
  fbm_node_t *B;
  B = A->right;
  fbm_set_right_child(A, B->left);
  fbm_replace_child_in_parent(T, A, B);
  fbm_set_left_child(B, A);
  fbm_calculate_height(A);
  fbm_calculate_height(B);
}

static void fbm_rotate_left_twolevels(fbm_t *T, fbm_node_t *A)
{
  fbm_node_t *B, *C;

  B = A->right; C = B->left;
  fbm_set_right_child(A, C->left);
  fbm_set_left_child(B, C->right);
  fbm_replace_child_in_parent(T, A, C);
  fbm_set_left_child(C, A);
  fbm_set_right_child(C, B);
  fbm_calculate_height(A);
  fbm_calculate_height(B);
  fbm_calculate_height(C);
}

static void fbm_rotate_right(fbm_t *T, fbm_node_t *A)
{
  fbm_node_t *B;
  B = A->left;
  fbm_set_left_child(A, B->right);
  fbm_replace_child_in_parent(T, A, B);
  fbm_set_right_child(B, A);
  fbm_calculate_height(A);
  fbm_calculate_height(B);
}

static void fbm_rotate_right_twolevels(fbm_t *T, fbm_node_t *A)
{
  fbm_node_t *B, *C;
  B = A->left; C = B->right;
  fbm_set_left_child(A, C->right);
  fbm_set_right_child(B, C->left);
  fbm_replace_child_in_parent(T, A, C);
  fbm_set_left_child(C, B);
  fbm_set_right_child(C, A);
  fbm_calculate_height(A);
  fbm_calculate_height(B);
  fbm_calculate_height(C);
}

static void fbm_rebalance(fbm_t *T, fbm_node_t *N)
{
  int u;
  fbm_node_t *P;

  fbm_calculate_height(N);

  u = fbm_HEIGHT(N->left)-fbm_HEIGHT(N->right);
  P = N->parent;
  if (u < -1) {
      if (fbm_left_heavy(N->right)) fbm_rotate_left_twolevels(T, N);
                                else fbm_rotate_left(T, N);
  } else if (u > 1) {
      if (fbm_right_heavy(N->left)) fbm_rotate_right_twolevels(T, N);
                                else fbm_rotate_right(T, N);
  }

  if (P) fbm_rebalance(T, P);
}

static fbm_node_t *fbm_remove_max(fbm_t *T,fbm_node_t *N)
{
   while(N->right) N= N->right;
   fbm_replace_child_in_parent(T, N, N->left);
   if (N->parent) fbm_rebalance(T,N->parent);
   return N;
}

static fbm_node_t *fbm_remove_min(fbm_t *T,fbm_node_t *N)
{
   while(N->left) N= N->left;
   fbm_replace_child_in_parent(T, N, N->right);
   if (N->parent) fbm_rebalance(T,N->parent);
   return N;
}

static void *fbm_remove_node(fbm_t *T, fbm_node_t *N)
{
   fbm_node_t *Z;
   void *R;
   if (!N->left && !N->right) {
     fbm_replace_child_in_parent(T, N, NULL);
     if (N->parent) fbm_rebalance(T,N->parent);
   } else {
     if (fbm_left_heavy(N)) Z= fbm_remove_max(T,N->left);
                        else Z= fbm_remove_min(T,N->right);
     fbm_replace_child_in_parent(T, N, Z);
     fbm_set_left_child(Z, N->left);
     fbm_set_right_child(Z, N->right);
     fbm_rebalance(T,Z);
   }
   R= N->data;
   free(N);
   return R;
}


static fbm_t *fbm_new()
{
  fbm_t *T;
  T= malloc(sizeof(*T));
  T->root= NULL;
  return T;
}

static free_block_t *fbm_find(fbm_t *T, uint64_t key)
{
   fbm_node_t *n;
   n= fbm_find_node(T, key);
   if (n) return n->data;
   return NULL;
}

static free_block_t* fbm_insert(fbm_t *T, free_block_t *data)
{
  fbm_node_t *N,*P,*K,**PL;
  int R;
  uint64_t lower; 
  uint64_t upper;

  lower= fbm_lower(data);
  upper= fbm_upper(data);
  if (fbm_compare(lower,upper)>0) return data;

  N = calloc(1, sizeof(*N));
  N->data= data;
  N->height= 1;

  P = 0;
  PL= &T->root;
  K= T->root;
  while (K) {
    R= fbm_compare(upper, fbm_lower(K->data));
    if (R==0) { free(N); return K->data; }
    if (R<0)  { PL= &K->left; P= K; K= K->left; continue; }
    R= fbm_compare(lower, fbm_upper(K->data));
    if (R>0) { PL= &K->right; P= K; K= K->right; continue; }
    else     { free(N); return K->data; }
  }
  *PL= N;
  N->parent= P;
  fbm_rebalance(T, N);
  return NULL;
}

static free_block_t *fbm_remove(fbm_t *T, uint64_t key)
{
   fbm_node_t *N;
   N= fbm_find_node(T,key);
   if (!N) return NULL;
   return fbm_remove_node(T,N);
}

static free_block_t *fbm_destroy(fbm_t *T)
{
   fbm_node_t *N= T->root;
   if (!N) return NULL;
   while(N->left) N=N->left;
   return fbm_remove_node(T,N);
}


typedef struct fbq_s fbq_t;


static fbq_t *fbq_new  (int initsize);
static void    fbq_free (fbq_t *q);
static void    fbq_put  (fbq_t *q, free_block_t* d);
static free_block_t*  fbq_get  (fbq_t *q);
static free_block_t*  fbq_peek (fbq_t *q);


static void fbq_remove(fbq_t *q, free_block_t* d);



struct fbq_s
{
  size_t len,siz;
  free_block_t* *data;
};


static int fbq_max(fbq_t *q, int idx1, int idx2)
{
  int C;
  free_block_t* A;
  free_block_t* B;
  A= q->data[idx1];
  B= q->data[idx2];
  C= (A->size<B->size?-1:(A->size>B->size?1:0));
  if (C>0) return idx1;
  return idx2;
}

static void fbq_exch(fbq_t *q,int a,int b)
{
  free_block_t* t;
  t= q->data[a];
  q->data[a]= q->data[b];
  q->data[b]= t;
  q->data[a]->qindex= a;
  q->data[b]->qindex= b;
}

static int fbq_bubble_up(fbq_t *q,int idx)
{
  int P,G,S;
  int v;
  v= 0;

again:
  if (idx==0) return v;

  P= (idx-1)/2;
  S= 2*P+1;
  if (S==idx) S++;
  if (S>=q->len) G= idx;
            else G= fbq_max(q, S, idx);

  if (fbq_max(q, P, G)==P) return v;
  fbq_exch(q,P,G);
  v++;
  idx= P;
  goto again;
}

static int fbq_bubble_down(fbq_t *q, int idx)
{
  int L,R,G;
  int v;
  v= 0;

again:
  L= 2*idx+1;
  if (L>=q->len) return v;
  R= L+1;
  if (R>=q->len) G= L;
            else G= fbq_max(q,L,R);
  if (fbq_max(q,G,idx)==idx) return v;
  fbq_exch(q,G,idx);
  idx= G;
  v++;
  goto again;
}


static fbq_t *fbq_new(int initsize)
{
  fbq_t *q;
  q= calloc(1,sizeof(*q));
  q->len= 0;
  if (initsize<2) initsize= 2;
  q->siz= initsize;
  q->data= malloc(sizeof(q->data[0])*q->siz);
  return q;
}

static void fbq_free(fbq_t *q)
{
  free(q->data);
  free(q);
}

static void fbq_put(fbq_t *q, free_block_t* d)
{
  if (q->len==q->siz)
  {
    q->siz+= q->siz/2;
    q->data= realloc(q->data,sizeof(q->data[0])*q->siz);
  }
  q->data[q->len]= d;
  d->qindex= q->len;
  q->len++;
  fbq_bubble_up(q, q->len-1);
}

static free_block_t* fbq_get(fbq_t *q)
{
  free_block_t* d;
  if (q->len==0) return NULL;
  d= q->data[0];
  q->data[0]= q->data[q->len-1];
  q->data[0]->qindex= 0;
  q->len--;
  fbq_bubble_down(q, 0);
  return d;
}

static free_block_t* fbq_peek(fbq_t *q)
{
  if (q->len==0) return NULL;
  return q->data[0];
}





static void fbq_remove(fbq_t *q, free_block_t* d)
{
  int K;
  K= d->qindex;
  if (K<0 || K>=q->len) return ;
  q->len--;
  if (K==q->len) return ;
  q->data[K]= q->data[q->len];
  q->data[K]->qindex= K;
  (void) (fbq_bubble_down(q, K) || fbq_bubble_up(q, K));
}




typedef struct { void *root; } qumap_t;


static void     *qumap_find(qumap_t *map,uint64_t key);
static void     *qumap_insert(qumap_t* map, uint64_t key, void *data);
static void     *qumap_remove(qumap_t* map, uint64_t key);
static void     *qumap_destroy(qumap_t *map);
static qumap_t  *qumap_new();
static void      qumap_free(qumap_t *map);


#ifndef _QUMAP_KEY
#define _QUMAP_KEY(X)    ((X)->d.leaf.key)
#define _QUMAP_DATA(X)   ((X)->d.leaf.data)
#define _QUMAP_CHLD(X,Y) ((X)->d.inte.child[Y])
#endif


typedef struct
{
  union { struct { void* child[2];           } inte;  
          struct { void *data; uint64_t key; } leaf; } d;
  unsigned internal:1, bit:6;
} qumap_node_t;


static inline int qumap_dir(uint64_t key,qumap_node_t *node)
{
  return (key >> node->bit)&1;
}


static inline qumap_node_t* qumap_leaf_node(uint64_t key, void *data)
{
  qumap_node_t *N;
  N= malloc(sizeof(*N));
  N->internal= 0;
  _QUMAP_KEY(N)= key;
  _QUMAP_DATA(N)= data; 
  return N;
}

static inline qumap_node_t* qumap_internal_node(uint32_t bit)
{
  qumap_node_t *N;
  N= malloc(sizeof(*N));
  N->internal= 1;
  N->bit= bit;
  return N;
}

static uint8_t qumap_crit_bit(uint64_t X, uint64_t Y)
{
  uint32_t i,j;
  static const uint64_t B[]=
  {
    0x5555555555555555ll,
    0x3333333333333333ll,
    0x0f0f0f0f0f0f0f0fll,
    0x00ff00ff00ff00ffll,
    0x0000ffff0000ffffll,
    0x00000000ffffffffll
  };
  uint64_t A;
  A= X ^ Y;
  
  for(i=1;i<64;i<<=1) A |= (A<<i);
  A = ~A;
  for(j=0,i=1;i<64;i<<=1,j++) A= (A&B[j]) + ((A>>i)&B[j]);
  return A;
}


static void* qumap_find(qumap_t *map,uint64_t key)
{
  qumap_node_t *node;

  node= map->root;
  if (!node) return NULL;

  while(node->internal) node= _QUMAP_CHLD(node, qumap_dir(key,node));

  if (_QUMAP_KEY(node)==key) return _QUMAP_DATA(node);
                          else return NULL;
}



static void* qumap_insert(qumap_t* map, uint64_t key, void *data)
{
  qumap_node_t *node_elt;

  if (!map->root)
  {
     map->root= qumap_leaf_node(key,data);
     return 0;
  }

  node_elt= map->root;
  while(node_elt->internal)
    node_elt= _QUMAP_CHLD(node_elt, qumap_dir(key,node_elt));

  
  uint64_t elt;
  int dir_elt;
  uint8_t bit;
  qumap_node_t *node_key;

  elt= _QUMAP_KEY(node_elt);

  if (elt==key) return _QUMAP_DATA(node_elt);

  bit= qumap_crit_bit(elt,key);

  dir_elt= (elt>>bit)&1;

  node_key= qumap_internal_node(bit);
  _QUMAP_CHLD(node_key,1-dir_elt)= qumap_leaf_node(key,data);

  qumap_node_t *parent;
  qumap_node_t *node;

  parent= 0;
  node= map->root;

  while(node->internal)
  {
     if (node->bit>bit) break;
     parent= node;
     node= _QUMAP_CHLD(node, qumap_dir(key,node));
  }

  _QUMAP_CHLD(node_key,dir_elt)= node;
  if (parent) { if (_QUMAP_CHLD(parent,0)==node)  _QUMAP_CHLD(parent,0)= node_key;
                                       else  _QUMAP_CHLD(parent,1)= node_key; }
  else map->root= node_key;
  return 0;
}

static void *qumap_remove(qumap_t* map, uint64_t key)
{
  void *retv;

  if (!map->root) return NULL;

   
  qumap_node_t *node;
  qumap_node_t *sibling;
  qumap_node_t *parent;
  qumap_node_t *gparent;

  node= map->root;
  gparent= 0;
  sibling= 0;
  parent= 0;

  while(node->internal)
  {
    int dir;
    dir= qumap_dir(key, node);
    gparent= parent;
    parent= node;
    sibling= _QUMAP_CHLD(node, 1-dir);
    node= _QUMAP_CHLD(node, dir);
  }

  if (_QUMAP_KEY(node)!=key) return NULL;
  retv= _QUMAP_DATA(node);
  free(node);

  if (!gparent)
  {
     map->root= sibling;
     if (parent) free(parent);
     return retv;
  }

  if (_QUMAP_CHLD(gparent,0)==parent) _QUMAP_CHLD(gparent,0)= sibling;
                            else _QUMAP_CHLD(gparent,1)= sibling;
  free(parent);

  return retv;
}

static void *qumap_destroy(qumap_t *map)
{
  qumap_node_t *node;
  qumap_node_t *parent;
  qumap_node_t *gparent;
  qumap_node_t *sibling;
  void *retv;

  if (!map->root) return 0;

  parent= 0;
  gparent= 0;
  sibling= 0;
  node= map->root;
  while(node->internal)
  {
    gparent= parent;
    parent= node;
    sibling= _QUMAP_CHLD(node, 1);
    node= _QUMAP_CHLD(node,0);
  }
  retv= _QUMAP_DATA(node);
  free(node);
  if (!gparent) 
  {
    map->root= sibling;
    if (parent) free(parent);
    return retv;
  }
  if (_QUMAP_CHLD(gparent,0)==parent)
    _QUMAP_CHLD(gparent,0)= sibling;
  else
    _QUMAP_CHLD(gparent,1)= sibling;

  free(parent);

  return retv;
}

static qumap_t  *qumap_new()
{
  qumap_t *s;
  s= malloc(sizeof(*s));
  s->root= NULL;
  return s;
}

static void qumap_free(qumap_t *map)
{
  free(map);
}

#line 367 "of.u"

struct _bff
{
  fbq_t *free_block_queue;
  fbm_t *free_block_map;
  qumap_t *object_map;
  uint64_t file_size;
  uint8_t magic[16];
  uint64_t next_avail_ident;

  uint8_t *buffer;
  uint64_t buffer_size;

  int mode;
  char *file_name;
  int file;
  int have_error;
  int eno;
#ifdef DEBUG
  uint64_t f_pos, f_buffer_size;
  uint8_t *f_buffer;
#endif

  uint32_t o_type;
  uint64_t o_size, o_len;
  uint8_t *o_buffer, *o_end, *o_pos;

  jmp_buf ejump;
  int n_ralloc, s_ralloc;
  void **ralloc;

  int w_pass;
};

#line 715 "of.test.c"

static int bff_read_file_header(bff_t *B);
static int bff_write_file_header(bff_t *B);
static int bff_write_block_header
    (bff_t *bf, uint64_t size, uint64_t prev, uint64_t next);
static int bff_read_block_header
    (bff_t *bf, uint64_t *size, uint64_t *prev, uint64_t *next);
static int bff_write_object_header
    (bff_t *bf, uint64_t size);
static int bff_read_object_header
    (bff_t *bf, uint64_t *size);
static int bff_write_object
   (bff_t *bf, uint64_t ident, uint64_t type, uint64_t osize, uint8_t *odata);
static int bff_write_object_eof
   (bff_t *bf, object_descriptor_t *O,uint64_t osize,  uint8_t *odata);
static int bff_write_object_single_free_block
       (bff_t *bf, object_descriptor_t *O,uint64_t osize,  uint8_t *odata, free_block_t *A);
static int bff_write_object_two_free_blocks
       (bff_t *bf, object_descriptor_t *O, uint64_t osize, uint8_t *odata,
        free_block_t *A, free_block_t *B);
static int bff_write_object_free_block_plus_eof
   (bff_t *bf, object_descriptor_t *O, 
    uint64_t osize, uint8_t *odata, free_block_t *A);
static int bff_free_object
   (bff_t *bf, uint64_t ident);
static void bff_record_free_block
       (bff_t *bf, free_block_t *F);
static void bff_record_object
       (bff_t *bf, object_descriptor_t *O);
static bff_t *bff_create(char *fn,char *magic);
static bff_t *bff_open(char *fn, int mode);
static void bff_close(bff_t *bf);
static bff_t *bff_new();
static int bff_write_bytes(bff_t *bf, uint64_t size, uint8_t *data);
static int bff_read_bytes(bff_t *bf, uint64_t size, uint8_t *data);
static int bff_write_formatted(bff_t *bf, const char *fmt, ...);
static int bff_read_formatted(bff_t *bf, const char *fmt, ...);
static int bff_seek(bff_t *bf, uint64_t pos);
static int bff_trunc(bff_t *bf, uint64_t size);
static int bff_write_junk(bff_t *bf, uint64_t size);
static int bff_read_index(bff_t *bf);
static void bff_index_entry
    (bff_t *bf, uint32_t type, uint32_t offset, uint32_t ident);
static int decode_unt32(uint8_t *d, uint32_t *R);
static int decode_unt64(uint8_t *d, uint64_t *R);
static int encode_unt64(uint8_t *d, uint64_t V);
static int encode_unt32(uint8_t *d, uint32_t V);
static int bff_write_index(bff_t *bf);
static int bff_read_object(bff_t *bf, uint64_t ident);
static void bff_fail_bread(bff_t *bf);
static int bff_bread_u8(bff_t *bf, uint8_t *R);
static int bff_bread_i8(bff_t *bf, int8_t *R) 
   ;
static int bff_bread_u16(bff_t *bf, uint16_t *R);
static int bff_bread_i16(bff_t *bf, int16_t *R) 
   ;
static int bff_bread_u32(bff_t *bf, uint32_t *R);
static int bff_bread_i32(bff_t *bf, int32_t *R) 
   ;
static int bff_bread_u64(bff_t *bf, uint64_t *R);
static int bff_bread_i64(bff_t *bf, int64_t *R) 
   ;
static int bff_bread_s8(bff_t *bf, uint8_t **R);
static int bff_bread_s16(bff_t *bf, uint16_t **R);
static int bff_bread_s32(bff_t *bf, uint32_t **R);
static void *bff_ralloc(bff_t *bf, uint32_t size);
static void *bff_ralloc0(bff_t *bf, uint32_t size);
static void bff_free_ralloc(bff_t *bf);
static void bff_reset_ralloc(bff_t *bf);
static void bff_write_start(bff_t *bf);
static void bff_bwrite_u8(bff_t *bf, uint8_t V);
static void bff_bwrite_u16(bff_t *bf, uint16_t V);
static void bff_bwrite_u32(bff_t *bf, uint32_t V);
static void bff_bwrite_u64(bff_t *bf, uint64_t V);
static void bff_bwrite_i8(bff_t *bf, int8_t V)
   ;
static void bff_bwrite_i16(bff_t *bf, int16_t V)
   ;
static void bff_bwrite_i32(bff_t *bf, int32_t V)
   ;
static void bff_bwrite_i64(bff_t *bf, int64_t V)
   ;
static void bff_bwrite_s8(bff_t *bf, uint8_t *D);
static void bff_bwrite_s16(bff_t *bf, uint16_t *D);
static void bff_bwrite_s32(bff_t *bf, uint32_t *D);
static int str16len(uint16_t *S);
static int str32len(uint32_t *S);
#line 39 "of.u"

static int bff_read_file_header(bff_t *B)
{
  if (bff_seek(B,0)) return 1;
     return bff_read_formatted(B, "bii", 16, 
              B->magic,&B->file_size, &B->next_avail_ident);
}

static int bff_write_file_header(bff_t *B)
{
  if (bff_seek(B,0)) return 1;
  return bff_write_formatted(B,"bii",16,
     B->magic, B->file_size, B->next_avail_ident);
}

static int bff_write_block_header
    (bff_t *bf, uint64_t size, uint64_t prev, uint64_t next)
{
  return bff_write_formatted(bf, "iii", size, prev, next);
}

static int bff_read_block_header
    (bff_t *bf, uint64_t *size, uint64_t *prev, uint64_t *next)
{
  return bff_read_formatted(bf, "iii", size, prev, next);
}

static int bff_write_object_header
    (bff_t *bf, uint64_t size)
{
  return bff_write_formatted(bf, "i", size);
}

static int bff_read_object_header
    (bff_t *bf, uint64_t *size)
{
  return bff_read_formatted(bf, "i", size);
}

static int bff_write_object
   (bff_t *bf, uint64_t ident, uint64_t type, uint64_t osize, uint8_t *odata)
{
  free_block_t *A, *B;
  uint64_t rsize;
  object_descriptor_t *O;

  O= malloc(sizeof(*O));
  O->type= type;
  O->ident= ident;

  A= fbq_get(bf->free_block_queue);
  if (!A) 
    return bff_write_object_eof(bf, O, osize, odata);

  rsize= osize+BLOCK_HEADER_SIZE+OBJECT_HEADER_SIZE;
  fbm_remove(bf->free_block_map, A->offset);
  if (A->size>=rsize)
    return bff_write_object_single_free_block(bf, O, osize, odata, A);

  rsize= osize+2*BLOCK_HEADER_SIZE+OBJECT_HEADER_SIZE;
  B= fbq_peek(bf->free_block_queue);
  if (!B || (A->size+B->size<rsize))
    return bff_write_object_free_block_plus_eof(bf, O, osize, odata,A);

  B= fbq_get(bf->free_block_queue);
  fbm_remove(bf->free_block_map, B->offset);
  return bff_write_object_two_free_blocks(bf, O, osize, odata,A,B);
}

static int bff_write_object_eof
   (bff_t *bf, object_descriptor_t *O,uint64_t osize,  uint8_t *odata)
{
  uint64_t bsize, offset, rsize;
  O->offset= bf->file_size;
  bsize= rsize= osize + BLOCK_HEADER_SIZE + OBJECT_HEADER_SIZE; 
  if (bsize<FREE_BLOCK_THRESHOLD)
     bsize= FREE_BLOCK_THRESHOLD;
  if (
    bff_seek(bf,bf->file_size) || 
    bff_write_block_header(bf, bsize, 0, 0) ||
    bff_write_object_header(bf, osize) ||
    bff_write_bytes(bf, osize, odata) ||
    bff_write_junk(bf, bsize-rsize) 
  )
     return 1;
  bf->file_size+= bsize;
  bff_record_object(bf, O);
  return 0;
}

static int bff_write_object_single_free_block
       (bff_t *bf, object_descriptor_t *O,uint64_t osize,  uint8_t *odata, free_block_t *A)
{
  uint64_t rsize,bsize;
  if (bff_seek(bf,A->offset)) return 1;
  O->offset= A->offset;
  rsize= osize + OBJECT_HEADER_SIZE + BLOCK_HEADER_SIZE;
  if (A->size-rsize<FREE_BLOCK_THRESHOLD)
    bsize= A->size; 
  else
    bsize= rsize;
  if (bff_write_block_header(bf, bsize, 0, 0) ||
      bff_write_object_header(bf, osize) ||
      bff_write_bytes(bf, osize, odata)) return 1;
  A->offset+= bsize;
  A->size-= bsize;
  if (A->size==0) 
     free(A);
  else
     bff_record_free_block(bf, A);
  bff_record_object(bf, O);
  return 0;
}

static int bff_write_object_two_free_blocks
       (bff_t *bf, object_descriptor_t *O, uint64_t osize, uint8_t *odata,
        free_block_t *A, free_block_t *B)
{
  // B is smaller than A. I'll use it completely, than put some
  // more into A. These variables record how much of the object
  // is put into each free block
  uint64_t osize_B, osize_A;
  // block sizes for each block
  uint64_t bsize_A, bsize_B;

  // these are the values to be used in sequence
  uint64_t OSIZ[2], bsize[2], offset[2];

  if (B->offset<A->offset)
     osize_B= B->size - OBJECT_HEADER_SIZE-BLOCK_HEADER_SIZE;
  else
     osize_B= B->size - BLOCK_HEADER_SIZE;

  if (osize_B>osize) osize_B= osize;

  osize_A= osize-osize_B;

  if (A->offset<B->offset)
    bsize_A= osize_A + OBJECT_HEADER_SIZE + BLOCK_HEADER_SIZE;
  else
    bsize_A= osize_A + BLOCK_HEADER_SIZE;

  if (A->size-bsize_A<FREE_BLOCK_THRESHOLD)
  {
     if (A->offset<B->offset)
     {
       osize_A+= A->size-bsize_A;
       osize_B-= A->size-bsize_A;
     }
     bsize_A= A->size;
  } 
  bsize_B= B->size;

  if (A->offset<B->offset)
  {
    OSIZ[0]= osize_A;
    bsize[0]= bsize_A;
    offset[0]= A->offset;
    OSIZ[1]= osize_B;
    bsize[1]= bsize_B;
    offset[1]= B->offset;
  }
  else
  {
    OSIZ[0]= osize_B;
    bsize[0]= bsize_B;
    offset[0]= B->offset;
    OSIZ[1]= osize_A;
    bsize[1]= bsize_A;
    offset[1]= A->offset;
  }

  if (
    bff_seek(bf, offset[0]) ||
    bff_write_block_header(bf, bsize[0], 0, offset[1]) ||
    bff_write_object_header(bf, osize) ||
    bff_write_bytes(bf, OSIZ[0], odata) || 
    bff_seek(bf, offset[1]) ||
    bff_write_block_header(bf, bsize[1], offset[0], 0) ||
    bff_write_bytes(bf, OSIZ[1], odata+OSIZ[0])
  )
      return 1;

  free(B);
  A->offset+= bsize_A;
  A->size-= bsize_A;
  if (A->size==0)
    free(A); 
  else
    bff_record_free_block(bf,A);
  O->offset= offset[0];
  bff_record_object(bf, O);
  return 0;
}

static int bff_write_object_free_block_plus_eof
   (bff_t *bf, object_descriptor_t *O, 
    uint64_t osize, uint8_t *odata, free_block_t *A)
{
  uint64_t offset2, offset1;
  uint64_t osize_A, osize_EOF;
  uint64_t bsize_A, bsize_EOF, rsize_EOF;

  osize_A= A->size - BLOCK_HEADER_SIZE - OBJECT_HEADER_SIZE;
  osize_EOF= osize - osize_A;

  bsize_A= A->size;
  rsize_EOF= bsize_EOF= osize_EOF + BLOCK_HEADER_SIZE;

  if (bsize_EOF < FREE_BLOCK_THRESHOLD)
    bsize_EOF= FREE_BLOCK_THRESHOLD;

  offset1= A->offset;
  offset2= bf->file_size;

  free(A);
  if (
    bff_seek(bf, offset1) ||
    bff_write_block_header(bf, bsize_A, 0, offset2) ||
    bff_write_object_header(bf, osize) ||
    bff_write_bytes(bf, osize_A, odata) ||
    bff_seek(bf, offset2) ||
    bff_write_block_header(bf, bsize_EOF, offset1, 0) ||
    bff_write_bytes(bf, osize_EOF, odata+osize_A) ||
    bff_write_junk(bf, bsize_EOF - rsize_EOF)
  )
     return 1;

  bf->file_size+= bsize_EOF;
  O->offset= offset1;
  bff_record_object(bf, O);
  return 0;
}

static int bff_free_object
   (bff_t *bf, uint64_t ident)
{
  free_block_t *F;
  uint64_t prev, next, size;
  object_descriptor_t *O;
  uint64_t bo;

  O= qumap_remove(bf->object_map, ident);
  if (!O) return 1;

  bo= O->offset;
  free(O);

  while(1)
  {
    if (bff_seek(bf, bo) ||
        bff_read_block_header(bf, &size, &prev, &next)) 
      return 1;
    F= calloc(1,sizeof(*F));
    F->offset= bo;
    F->size= size;
    bff_record_free_block(bf,F);
    bo= next;
    if (!bo) break;
  }

  return 0;
}

static void bff_record_free_block
       (bff_t *bf, free_block_t *F)
{
  free_block_t *A,*B;

  A= fbm_remove(bf->free_block_map, F->offset-1);
  B= fbm_remove(bf->free_block_map, F->offset+F->size);

  if (A) { F->offset= A->offset; F->size+= A->size;
           fbq_remove(bf->free_block_queue,A); free(A); }
  if (B) { F->size+= B->size; 
           fbq_remove(bf->free_block_queue, B); free(B); }

  if (F->offset+F->size==bf->file_size)
  {
    bf->file_size= F->offset;
    free(F);
    return ;
  }
  fbm_insert(bf->free_block_map, F);
  fbq_put(bf->free_block_queue, F);
}

static void bff_record_object
       (bff_t *bf, object_descriptor_t *O)
{
  qumap_insert(bf->object_map, O->ident, O);
}

#line 409 "of.u"

static bff_t *bff_create(char *fn,char *magic)
{
  bff_t *bf;
  bf= bff_new();
  bf->file_name= strdup(fn);
  memcpy(bf->magic, magic, 16);
  bf->file= open(fn, O_RDWR | O_CREAT | O_TRUNC, 0600);
  if (bf->file==-1) 
  {
    bf->have_error= 1;
    bf->eno= errno;
  }
  bf->mode= 'W';
  bff_write_file_header(bf);
  bff_write_junk(bf, bf->file_size - FILE_HEADER_SIZE);
  return bf;
}

static bff_t *bff_open(char *fn, int mode)
{
  bff_t *bf;
  bf= bff_new();
  bf->file_name= strdup(fn);
  bf->file= open(fn, O_RDWR);
  if (bf->file==-1) 
  {
    bf->have_error= 1;
    bf->eno= errno;
  }
  bf->mode= mode;
  bff_read_file_header(bf);
  bff_read_index(bf);
  return bf;
}

static void bff_close(bff_t *bf)
{
  object_descriptor_t *O;
  free_block_t *F;
  if (bf->mode=='W')
    bff_write_index(bf);

  while((F=fbm_destroy(bf->free_block_map))) ;
  while((F=fbq_get(bf->free_block_queue)))
    free(F);
  fbq_free(bf->free_block_queue);
  while((O=qumap_destroy(bf->object_map)))
    free(O);
  qumap_free(bf->object_map);
  
  if (bf->buffer) free(bf->buffer);
  if (bf->file_name) free(bf->file_name);
  if (bf->file>=0) close(bf->file);
  bff_free_ralloc(bf);
  if (bf->ralloc) free(bf->ralloc);
  memset(bf, 0, sizeof(*bf));
  free(bf);
}

static bff_t *bff_new()
{
  bff_t *bf;
  bf= calloc(1,sizeof(*bf));
  bf->free_block_queue= fbq_new(100);
  bf->free_block_map= fbm_new();
  bf->object_map= qumap_new();
  bf->buffer_size= 1000;
  bf->buffer= malloc(bf->buffer_size);

  bf->next_avail_ident= 32;

  // reserve this for the file header
  bf->file_size= 128;
#ifdef DEBUG
  bf->f_buffer_size= 1000;
  bf->f_buffer= malloc(bf->f_buffer_size);
  bf->f_pos= 0;
#endif
  return bf;
}

static int bff_write_bytes(bff_t *bf, uint64_t size, uint8_t *data)
{
  if (!size) return 0;
#ifdef DEBUG
  if (bf->f_pos+size>bf->f_buffer_size)
  {
    bf->f_buffer_size+= bf->f_buffer_size/2+size;
    bf->f_buffer= realloc(bf->f_buffer, bf->f_buffer_size);
  }
  memcpy(bf->f_buffer+bf->f_pos, data, size);
  bf->f_pos+= size;
  return 0;
#endif
  if (bf->file<0 || bf->mode=='R' || bf->have_error) return 1;
  ssize_t K;
  while(size)
  {
    K= write(bf->file, data, size);
    if (K<=0)
    {
      if (errno==EINTR) continue;
      bf->eno= errno;
      bf->have_error= 1;
      return 1;
    }
    data+= K;
    size-= K;
  }
  return 0;
}

static int bff_read_bytes(bff_t *bf, uint64_t size, uint8_t *data)
{
  if (size==0) return 0;
#ifdef DEBUG
  if (bf->f_pos+size>bf->f_buffer_size) return 1;
  memcpy(data, bf->f_buffer+bf->f_pos, size);
  bf->f_pos+= size;
  return 0;
#endif
  if (bf->file<0 || bf->have_error) return 1;
  ssize_t K;
  while(size)
  {
    K= read(bf->file, data, size);
    if (K<=0)
    {
      if (errno==EINTR) continue;
      bf->eno= errno;
      bf->have_error= 1;
      return 1;
    }
    data+= K;
    size-= K;
  }
  return 0;
}

static int bff_write_formatted(bff_t *bf, const char *fmt, ...)
{
  va_list ap;
  int sz;
  uint8_t *k; uint64_t M;
  int i,j,b;
  va_start(ap,fmt);
  sz= 0;
  for(i=0;fmt[i];i++)
   switch(fmt[i])
   { case 'b': sz+= va_arg(ap, int); break;
     case 'i': sz+= 8; va_arg(ap, uint64_t); break; }
  va_end(ap);
  if (bf->buffer_size<sz) 
  {
    bf->buffer_size= sz;
    bf->buffer= realloc(bf->buffer, bf->buffer_size);
  }
  sz= 0;
  va_start(ap,fmt);
  for(i=0;fmt[i];i++)
   switch(fmt[i])
   { case 'b': b= va_arg(ap, int); k= va_arg(ap, uint8_t*);
               memcpy(bf->buffer+sz, k, b);
               sz+= b;
               break;
     case 'i': M= va_arg(ap, uint64_t);
               for(j=0;j<8;j++) { bf->buffer[sz+j]= M; M>>= 8; }
               sz+= 8;
               break; }
  va_end(ap);
  return bff_write_bytes(bf, sz, bf->buffer);
}

static int bff_read_formatted(bff_t *bf, const char *fmt, ...)
{
  va_list ap;
  int sz;
  uint8_t *k; uint64_t *M,V,U;
  int i,j,b;
  va_start(ap,fmt);
  sz= 0;
  for(i=0;fmt[i];i++)
   switch(fmt[i])
   { case 'b': sz+= va_arg(ap, int); break;
     case 'i': sz+= 8; va_arg(ap, uint64_t); break; }
  va_end(ap);
  if (bf->buffer_size<sz) 
  {
    bf->buffer_size= sz;
    bf->buffer= realloc(bf->buffer, bf->buffer_size);
  }
  if (bff_read_bytes(bf, sz, bf->buffer)) return 1;
  sz= 0;
  va_start(ap,fmt);
  for(i=0;fmt[i];i++)
   switch(fmt[i])
   { case 'b': b= va_arg(ap, int); k= va_arg(ap, uint8_t*);
               memcpy(k, bf->buffer+sz, b);
               sz+= b;
               break;
     case 'i': M= va_arg(ap, uint64_t*);
               V= 0; 
               for(j=0;j<8;j++) { U= bf->buffer[sz+j]; V|= U<<(j*8); }
               *M= V;
               sz+= 8;
               break; }
  va_end(ap);
  return 0;
}


static int bff_seek(bff_t *bf, uint64_t pos)
{
#ifdef DEBUG
  if (pos>bf->f_buffer_size)
  {
    bf->f_buffer_size= pos + 1000;
    bf->f_buffer= realloc(bf->f_buffer, bf->f_buffer_size);
  }
  bf->f_pos= pos;
  return 0;
#endif
  if (bf->file<0 || bf->have_error) return 1;
  if (lseek(bf->file,pos,SEEK_SET)==(off_t)-1)
  {
    bf->eno= errno;
    bf->have_error= 1;
    return 1;
  }
  return 0;
}

static int bff_trunc(bff_t *bf, uint64_t size)
{
  if (bf->have_error || bf->file<0) return 1;
  if (ftruncate(bf->file, size)) return 1;
  return 0;
}

static int bff_write_junk(bff_t *bf, uint64_t size)
{
  uint8_t g[200];
  uint64_t cp;
  if (size==0) return 0;
  while(size)
  {
    cp= size<sizeof(g) ? size : sizeof(g);
    if (bff_write_bytes(bf, cp, g)) return 1;
    size-= cp;
  }
  return 0;
}

#line 689 "of.u"

static int bff_read_index(bff_t *bf)
{
  uint8_t data[20*100], *d;
  uint32_t N, i, cp;
  uint32_t T; uint64_t O,I;
  if (bff_seek(bf, bf->file_size)) return 1;
  if (bff_read_bytes(bf, 4, data)) return 1;
  decode_unt32(data,&N);
  while(N)
  {
    cp= 100;
    if (cp>N) cp= N;
    if (bff_read_bytes(bf, cp*20, data)) return 1;
    d= data;
    for(i=0;i<cp;i++)
    {
      d+= decode_unt32(d, &T);
      d+= decode_unt64(d, &O);
      d+= decode_unt64(d, &I);
      bff_index_entry(bf, T, O, I);
    }
    N-= cp;
  }
  return 0;
}

static void bff_index_entry
    (bff_t *bf, uint32_t type, uint32_t offset, uint32_t ident)
{
  if (type==0)
  {
    free_block_t *F;
    F= calloc(1,sizeof(*F));
    F->offset= offset;
    F->size= ident;
    fbm_insert(bf->free_block_map, F);
    fbq_put(bf->free_block_queue, F);
  } 
  else
  {
    object_descriptor_t *O; 
    O= calloc(1,sizeof(*O));
    O->ident= ident;
    O->type= type;
    O->offset= offset;
    qumap_insert(bf->object_map, O->ident, O);
  }
}

static int decode_unt32(uint8_t *d, uint32_t *R)
{
  uint32_t V,L,i;
  V= 0;
  for(i=0;i<4;i++)
  {
     L= d[i];
     V|= L << (i*8);
  }
  *R= V;
  return 4;
}

static int decode_unt64(uint8_t *d, uint64_t *R)
{
  uint64_t V,L,i;
  V= 0;
  for(i=0;i<8;i++)
  {
     L= d[i];
     V|= L << (i*8);
  }
  *R= V;
  return 8;
}

static int encode_unt64(uint8_t *d, uint64_t V)
{
  int i;
  for(i=0;i<8;i++)
  {
    d[i]= V;
    V>>= 8;
  }
  return 8;
}

static int encode_unt32(uint8_t *d, uint32_t V)
{
  int i;
  for(i=0;i<4;i++)
  {
    d[i]= V;
    V>>= 8;
  }
  return 4;
}

static int bff_write_index(bff_t *bf)
{
  object_descriptor_t *O;
  free_block_t *F;
  uint32_t N;
  uint8_t d[20*100];
  uint8_t *data,*end;

  if (bf->have_error || bf->mode=='R') return 1;
  if (bff_seek(bf, bf->file_size) || bff_write_junk(bf, 4)) return 1;

  end= d + sizeof(d);
  data= d;
  N= 0;

  while((F=fbm_destroy(bf->free_block_map)))
  {
    data+= encode_unt32(data, 0);
    data+= encode_unt64(data, F->offset);
    data+= encode_unt64(data, F->size);
    N++;
    if (data==end)
    {
       if (bff_write_bytes(bf, data-d, data)) return 1;
       data= d;
    }
  }

  while((O=qumap_destroy(bf->object_map)))
  {
    data+= encode_unt32(data, O->type);
    data+= encode_unt64(data, O->offset);
    data+= encode_unt64(data, O->ident);
    N++;
    free(O);
    if (data==end)
    {
       if (bff_write_bytes(bf, data-d, data)) return 1;
       data= d;
    }
  }

  if (data!=d && bff_write_bytes(bf, data-d, data)) return 1;
  
  if (bff_seek(bf, bf->file_size) || 
      bff_write_formatted(bf,"i",N)) 
    return 1;

  if (bff_trunc(bf, bf->file_size + 4 + 20*N)) return 1;
  return 0;
}

static int bff_read_object(bff_t *bf, uint64_t ident)
{
  object_descriptor_t *O;
  uint8_t *data;
  uint64_t osize, bsize, bprev, bnext, rsize, hdr_size;

  O= qumap_find(bf->object_map, ident);
  if (!O) return 1;
  bf->o_type= O->type;
  if (bff_seek(bf, O->offset)) return 1;
  if (bff_read_block_header(bf, &bsize, &bprev, &bnext))
     return 1;
  if (bff_read_object_header(bf, &osize)) return 1;
  bf->o_len= osize;
  if (bf->o_size < osize)
  {
    bf->o_size+= osize + bf->o_size/2;
    if (bf->o_buffer)
       bf->o_buffer= realloc(bf->o_buffer, bf->o_size);
    else
       bf->o_buffer= malloc(bf->o_size);
  }
  bf->o_end= bf->o_buffer+ bf->o_len;
  bf->o_pos= bf->o_buffer;
  data= bf->o_buffer;
  // printf("object size: %d. reading ", (int) osize);
  hdr_size= BLOCK_HEADER_SIZE+OBJECT_HEADER_SIZE;
  while(1)
  {
    rsize= bsize - hdr_size;
    if (rsize>osize) rsize= osize;
    // printf(" %d ", (int) rsize);
    if (bff_read_bytes(bf, rsize, data)) return 1;
    osize-= rsize;
    data+= rsize;
    if (osize==0) break;
    if (bff_seek(bf, bnext) || 
        bff_read_block_header(bf, &bsize, &bprev, &bnext)) return 1;
    hdr_size= BLOCK_HEADER_SIZE;
  }
 // printf("\n");
  return 0;
}

#line 887 "of.u"

static void bff_fail_bread(bff_t *bf)
{
  bff_free_ralloc(bf);
  longjmp(bf->ejump, 1);
}

static int bff_bread_u8(bff_t *bf, uint8_t *R)
{
  if (bf->o_pos==bf->o_end) bff_fail_bread(bf);
  *R= *bf->o_pos++;
  return 0;
}

static int bff_bread_i8(bff_t *bf, int8_t *R) 
   { return bff_bread_u8(bf, (uint8_t*) R); }

static int bff_bread_u16(bff_t *bf, uint16_t *R)
{
  uint16_t V,D;
  if (bf->o_pos+2>bf->o_end) bff_fail_bread(bf);
  V= bf->o_pos[0];
  D= bf->o_pos[1];
  *R = V | (D<<8);
  bf->o_pos+= 2;
  return 0;
}

static int bff_bread_i16(bff_t *bf, int16_t *R) 
   { return bff_bread_u16(bf, (uint16_t*) R); }

static int bff_bread_u32(bff_t *bf, uint32_t *R)
{
  uint32_t V,D,i;
  if (bf->o_pos+4>bf->o_end) bff_fail_bread(bf);
  V= 0;
  for(i=0;i<4;i++) { D= bf->o_pos[i]; V |= D<<(i*8); }
  *R= V;
  bf->o_pos+= 4;
  return 0;
}

static int bff_bread_i32(bff_t *bf, int32_t *R) 
   { return bff_bread_u32(bf, (uint32_t*) R); }


static int bff_bread_u64(bff_t *bf, uint64_t *R)
{
  uint64_t V,D,i;
  if (bf->o_pos+8>bf->o_end) bff_fail_bread(bf);
  V= 0;
  for(i=0;i<8;i++) { D= bf->o_pos[i]; V |= D<<(i*8); }
  *R= V;
  bf->o_pos+= 8;
  return 0;
}

static int bff_bread_i64(bff_t *bf, int64_t *R) 
   { return bff_bread_u64(bf, (uint64_t*) R); }

static int bff_bread_s8(bff_t *bf, uint8_t **R)
{
  uint8_t *D;
  uint32_t N;
  if (bff_bread_u32(bf, &N)) return 1;
  D= bff_ralloc(bf,N+1);
  if (bf->o_pos+N>bf->o_end) bff_fail_bread(bf);
  memcpy(D, bf->o_pos, N);
  D[N]= 0;
  *R= D;
  bf->o_pos+= N;
  return 0;
}

static int bff_bread_s16(bff_t *bf, uint16_t **R)
{
  uint16_t *D;
  uint32_t N,i;
  if (bff_bread_u32(bf, &N)) return 1;
  D= bff_ralloc(bf,(N+1)*sizeof(D[0]));
  if (bf->o_pos+2*N>bf->o_end) bff_fail_bread(bf);
  for(i=0;i<N;i++)
  {
    D[i]= bf->o_pos[0] | (bf->o_pos[1]<<8);
    bf->o_pos+= 2;
  }
  D[N]= 0;
  *R= D;
  return 0;
}

static int bff_bread_s32(bff_t *bf, uint32_t **R)
{
  uint32_t *D;
  uint32_t N,i;
  if (bff_bread_u32(bf, &N)) return 1;
  D= bff_ralloc(bf,(N+1)*sizeof(D[0]));
  if (bf->o_pos+4*N>bf->o_end) bff_fail_bread(bf);
  for(i=0;i<N;i++)
  {
    D[i]= bf->o_pos[0] 
          | ((uint32_t)bf->o_pos[1]<<8) 
          | ((uint32_t)bf->o_pos[2]<<16) 
          | ((uint32_t)bf->o_pos[3]<<24);
    bf->o_pos+= 4;
  }
  D[N]= 0;
  *R= D;
  return 0;
}

#line 1020 "of.u"

static void *bff_ralloc(bff_t *bf, uint32_t size)
{
  void *R;
  if (bf->n_ralloc==bf->s_ralloc)
  {
    bf->s_ralloc+= bf->s_ralloc/2 + 50;
    if (bf->ralloc) 
      bf->ralloc= realloc(bf->ralloc, bf->s_ralloc*sizeof(bf->ralloc[0]));
    else
      bf->ralloc= malloc(bf->s_ralloc*sizeof(bf->ralloc[0]));
  }
  R= bf->ralloc[bf->n_ralloc++]= malloc(size);
  return R;
}

static void *bff_ralloc0(bff_t *bf, uint32_t size)
{
  void *R;
  R= bff_ralloc(bf,size);
  memset(R, 0, size);
  return R;
}

static void bff_free_ralloc(bff_t *bf)
{
  int i;
  for(i=0;i<bf->n_ralloc;i++)
   free(bf->ralloc[i]);
  bf->n_ralloc= 0;
}

static void bff_reset_ralloc(bff_t *bf)
{
  bf->n_ralloc= 0;
}

#line 1078 "of.u"

static void bff_write_start(bff_t *bf)
{
  bf->o_len= 0;
}

static void bff_bwrite_u8(bff_t *bf, uint8_t V)
{
  if (bf->w_pass)
  {
    bf->o_pos[0]= V;
  }
  bf->o_pos++;
}

static void bff_bwrite_u16(bff_t *bf, uint16_t V)
{
  if (bf->w_pass)
  {
    bf->o_pos[0]= V;
    bf->o_pos[1]= V>>8;
  }
  bf->o_pos+= 2;
}

static void bff_bwrite_u32(bff_t *bf, uint32_t V)
{
  if (bf->w_pass)
  {
    bf->o_pos[0]= V;
    bf->o_pos[1]= V>>8;
    bf->o_pos[2]= V>>16;
    bf->o_pos[3]= V>>24;
  }
  bf->o_pos+= 4;
}

static void bff_bwrite_u64(bff_t *bf, uint64_t V)
{
  int i;
  if (bf->w_pass)
  {
    for(i=0;i<8;i++)
    {
      bf->o_pos[i]= V;
      V>>= 8;
    }
  }
  bf->o_pos+= 8;
}

static void bff_bwrite_i8(bff_t *bf, int8_t V)
   { bff_bwrite_u8(bf,(uint8_t)V); }

static void bff_bwrite_i16(bff_t *bf, int16_t V)
   { bff_bwrite_u16(bf,(uint16_t)V); }

static void bff_bwrite_i32(bff_t *bf, int32_t V)
   { bff_bwrite_u32(bf,(uint32_t)V); }

static void bff_bwrite_i64(bff_t *bf, int64_t V)
   { bff_bwrite_u64(bf,(uint64_t)V); }


static void bff_bwrite_s8(bff_t *bf, uint8_t *D)
{
  uint32_t i,L;
  L= D ? 0 : strlen((char*) D);
  bff_bwrite_u32(bf, L);
  if (bf->w_pass)
  {
    for(i=0;i<L;i++) bf->o_pos[i]= D[i];
  }
  bf->o_pos+= L;
}

static void bff_bwrite_s16(bff_t *bf, uint16_t *D)
{
  uint32_t i,L;
  L= D ? 0 : str16len(D);
  bff_bwrite_u32(bf, L);
  if (bf->w_pass)
  {
    for(i=0;i<L;i++) 
    {
      bf->o_pos[2*i+0]= D[i]; 
      bf->o_pos[2*i+1]= D[i]>>8; 
    }
  }
  bf->o_pos+= 2*L;
}

static void bff_bwrite_s32(bff_t *bf, uint32_t *D)
{
  uint32_t i,L;
  L= D ? 0 : str32len(D);
  bff_bwrite_u32(bf, L);
  if (bf->w_pass)
  {
    for(i=0;i<L;i++) 
    {
      bf->o_pos[4*i+0]= D[i]; 
      bf->o_pos[4*i+1]= D[i]>>8; 
      bf->o_pos[4*i+2]= D[i]>>16; 
      bf->o_pos[4*i+3]= D[i]>>24; 
    }
  }
  bf->o_pos+= 4*L;
}

static int str16len(uint16_t *S)
{
  int i;
  for(i=0;S[i];i++) ;
  return i;
}

static int str32len(uint32_t *S)
{
  int i;
  for(i=0;S[i];i++) ;
  return i;
}

#line 17 "of.test.u"

typedef struct { uint64_t offset, size; } blo_t;

#line 1825 "of.test.c"

typedef struct blb_buffer blb_t;

static blb_t *blb_new     (size_t blksize);
static void    blb_free    (blb_t *B, blo_t **Rd, size_t *Rl);
static void    blb_append  (blb_t *dst, blb_t *src_f);
static void    blb_put     (blb_t *dst, blo_t d);
static void    blb_write   (blb_t *B, blo_t *d, size_t l);
static size_t  blb_read    (blb_t *B, blo_t *d, size_t l);
static void    blb_collect (blb_t *B, blo_t **Rd, size_t *Rl);

struct blb_node { 
  struct blb_node *next;
  size_t len,size,start;
  blo_t data[1];
};

struct blb_buffer {
  struct blb_node *first, *last;
  size_t blksize;
};



static struct blb_node *blb_get_free_node(blb_t *B)
{
  struct blb_node *P;
  
  P= B->last;
  if (P && (P->start+P->len)<P->size) 
    return P;

  P= malloc(sizeof(*P)+(B->blksize-1)*sizeof(blo_t));
  P->len= 0;
  P->next= NULL;
  P->size= B->blksize;
  P->start= 0;
  if (!B->last) { B->last= B->first= P; }
           else { B->last->next= P; B->last= P; }
  return P;
}


static blb_t *blb_new(size_t blksize)
{
  blb_t *B;
  B= malloc(sizeof(*B));
  B->first= B->last= NULL;
  B->blksize= blksize;
  return B;
}


static void blb_free(blb_t *B, blo_t **Rd, size_t *Rl)
{
  struct blb_node *P,*N;
  if (Rd || Rl) blb_collect(B, Rd, Rl);
  for(P=B->first;P;P=N)
  {
    N= P->next;
    free(P);
  }
  free(B);
}

static void blb_collect(blb_t *B, blo_t **Rd, size_t *Rl)
{
  size_t S;
  struct blb_node *P;
  blo_t *R;

  if (!Rd && !Rl) return ;

  S= 0;
  for(P=B->first;P;P=P->next) S+= P->len;

  if (S==0) { if (Rd) *Rd= NULL; if (Rl) *Rl= 0; return ; }
  if (!Rd) { *Rl= S; return ; }

  R= malloc(S*sizeof(blo_t));
  S= 0;
  for(P=B->first;P;P=P->next)
  {
    memcpy(R+S, P->data+P->start, P->len*sizeof(blo_t));
    S+= P->len;
  }

  *Rd= (void*) R;
  if (Rl) *Rl= S;
}

static void blb_write(blb_t *B, blo_t *d, size_t l)
{
  struct blb_node *P;
  size_t cp;

  while(l)
  {
    P= blb_get_free_node(B);
    cp= P->size-P->len-P->start;
    if (cp>l) cp= l; 

    memcpy(P->data+P->len+P->start, d, cp*sizeof(blo_t));
    P->len+= cp;
    l-= cp;
    d+= cp;
  }
}

static void blb_put(blb_t *B, blo_t d)
{
  struct blb_node *P;
  P= blb_get_free_node(B);
  P->data[P->start+P->len++]= d;
}

static void blb_append(blb_t *dst, blb_t *src_f)
{
  if (src_f->first!=NULL)
  {
    if (dst->first==NULL)
    {
      dst->first= src_f->first;
      dst->last= src_f->last;
    }
    else
    {
      dst->last->next= src_f->first;
      dst->last= src_f->last;
    }
  }
  free(src_f);
}

static size_t blb_read(blb_t *B, blo_t *d, size_t l)
{
  size_t total;
  struct blb_node *P;
  size_t cp;
  total= 0;

  while(l)
  {
    P= B->first;
    if (!P) break;  
    if (P->len==0)
    {
      B->first= P->next;
      if (!B->first) B->last= NULL;
      free(P);
      continue;
    } 
    cp= P->len;
    if (cp>l) cp= l;
    memcpy(d, P->data+P->start, cp*sizeof(blo_t));
    P->start+= cp;
    P->len-= cp;
    total+= cp;
    d+= cp;
  }
  return total;
}

#line 21 "of.test.u"

void seed_random(int *argc,char **argv)
{
  unsigned long int seed;
  int i;
  for(i=1;i<*argc-1;i++)
    if (!strcmp(argv[i],"-s"))
    {
      seed= strtoul(argv[i+1], NULL,0);
      for(;i<*argc-2;i++) argv[i]= argv[i+2];
      (*argc)--;
      goto done;
    }
  seed= time(NULL); 
done:
  printf("seed is %#lX\n", seed); 
  srand(seed);
}

static int obsize;
static int maxnob;
typedef struct { uint64_t ident, size; } tob_t;
static tob_t *Tob;
static int nTob,sTob;
#define MAX_TESTOBJ_SIZE 2048
#define MIN_TESTOBJ_SIZE 20
static uint8_t udata[MAX_TESTOBJ_SIZE];

static int Ni, Nd, Nf;

void fail() { *(int*)0= 0; }

static unsigned int ubyte(uint64_t ident, int o)
{
#if 0
  int sh;
  uint64_t V;
  sh= o/8;
  V= (ident<<sh);
  if (V==0) V= ident;
  return (V>>((sh%8)*8))&0xff;
#else
  return (o*o)&0xff;
#endif
}

static void prepare_udata(uint64_t ident, int size)
{
  int i;
  for(i=0;i<size;i++)
    udata[i]= ubyte(ident,i);
}

static void check_udata(uint64_t ident, int size, uint8_t *oda)
{
  int i;
  for(i=0;i<size;i++)
    if (oda[i]!=ubyte(ident,i)) { printf("failed at byte %d\n",i); fail(); }
}

static void do_insert(bff_t *bf)
{
  if (nTob==sTob)  
  {
    if (!sTob) { sTob= 200; Tob= malloc(sizeof(*Tob)*sTob); }
    else { sTob+= sTob/2; Tob= realloc(Tob, sizeof(*Tob)*sTob); }
  }
  Tob[nTob].ident= bf->next_avail_ident++;
  Tob[nTob].size= rand()%(MAX_TESTOBJ_SIZE-MIN_TESTOBJ_SIZE)+MIN_TESTOBJ_SIZE;

  prepare_udata(Tob[nTob].ident, Tob[nTob].size);
 
  if (bff_write_object(bf, Tob[nTob].ident, 50, Tob[nTob].size, udata)) fail(); 
  obsize+= Tob[nTob].size;
  nTob++;
  if (nTob>maxnob) maxnob= nTob;
  Ni++;
}

static void do_remove(bff_t *bf)
{
  int K;
  tob_t T;
  if (!nTob) return ;
  K= rand() % nTob;
  T= Tob[K];
  Tob[K]= Tob[nTob-1];
  nTob--;
  if (bff_free_object(bf, T.ident)) fail();
  obsize-= T.size;
  Nd++;
}

static void dump_objects(bff_t *bf)
{
  uint64_t bo, size, prev, next;
  object_descriptor_t *O;
  qumap_t *OM;
  OM= qumap_new();
  while((O=qumap_destroy(bf->object_map)))
  {
    qumap_insert(OM, O->ident, O);
    printf("O %#llx : ", (unsigned long long int) O->offset); 
                                            /// , (int) O->size);
    bo= O->offset;
    do {
      if (bff_seek(bf, bo)) { printf("\n"); return ; }
      bff_read_block_header(bf, &size, &prev, &bo);
      printf(" %d", (int) size);
    } while(bo);
    printf("\n");
  }
  bf->object_map= OM;
}

static void dump_free(bff_t *bf)
{
  uint64_t K;
  free_block_t *F;
  fbm_t *FM;
  FM= fbm_new();
  
  while((F=fbm_destroy(bf->free_block_map)))
  {
    printf("F: %#llx - %d\n", (unsigned long long int) F->offset, (int) F->size);
    fbm_insert(FM, F);
  }
  bf->free_block_map= FM;
}

static void fail_check()
{
  fail();
}

static int blosizcmp(const void *_a, const void *_b)
{
  const blo_t *a= _a, *b= _b;
  if (a->size < b->size ) return -1;
  if (a->size > b->size ) return 1;
  return 0;
}


static int blocmp(const void *_a, const void *_b)
{
  const blo_t *a= _a, *b= _b;
  if (a->offset < b->offset ) return -1;
  if (a->offset > b->offset ) return 1;
  return 0;
}

static void check_file(bff_t *bf)
{
  fbm_t *FM;
  free_block_t *F;
  qumap_t *OM;
  blb_t *bl;
  uint64_t bo,prev;
  object_descriptor_t *O;
  blo_t *A; size_t nA; int i;
  
  FM= fbm_new();
  OM= qumap_new();
  bl= blb_new(100);

  while((O=qumap_destroy(bf->object_map)))
  {
    qumap_insert(OM, O->ident, O);
    bo= O->offset;
    do
    {
      blo_t B;
      B.offset= bo;
      bff_seek(bf, bo);
      bff_read_block_header(bf, &B.size, &prev, &bo);
      blb_put(bl, B);
    } while (bo);
  }
  bf->object_map= OM;

  while((F=fbm_destroy(bf->free_block_map)))
  {
    fbm_insert(FM, F);
    blo_t B;
    B.offset= F->offset;
    B.size= F->size;
    blb_put(bl, B);
  }
  bf->free_block_map= FM;

  blb_free(bl, &A, &nA);
  if (nA>1)
  {
    qsort(A, nA, sizeof(A[0]), blocmp);
    for(i=0;i<nA-1;i++)
      if (A[i].offset+A[i].size!=A[i+1].offset)
        fail_check();
  }
  if (A) free(A);
}

static void do_find(bff_t *bf)
{
  int K;

  if (!nTob) return ;
  K= rand() % nTob;
  if (bff_read_object(bf, Tob[K].ident)) fail();

  if (bf->o_size!=Tob[K].size) fail();
  check_udata(Tob[K].ident, bf->o_size, bf->o_buffer);
  Nf++;
}

static void analyse(bff_t *bf)
{
  free_block_t *F;
  blb_t *bl;
  uint64_t bo,prev;
  object_descriptor_t *O;
  blo_t *A; size_t nA; int i;
  FILE *fA;
    int j,c;
  int p;
  fA= fopen("bsize.txt", "w");
  
  bl= blb_new(100);

  while((O=qumap_destroy(bf->object_map)))
  {
    bo= O->offset;
    do
    {
      blo_t B;
      B.offset= bo;
      bff_seek(bf, bo);
      bff_read_block_header(bf, &B.size, &prev, &bo);
      blb_put(bl, B);
    } while (bo);
  }

  while((F=fbm_destroy(bf->free_block_map)))
  {
    blo_t B;
    B.offset= F->offset;
    B.size= F->size;
    blb_put(bl, B);
  }

  blb_free(bl, &A, &nA);
  if (nA>1)
  {
    qsort(A, nA, sizeof(A[0]), blosizcmp);
    j= 0; c=1;
    p= 0;
    for(i=1;i<nA;i++)
    {
      if (A[i].size==A[j].size) { c++; continue; }
      fprintf(fA, "%d %d\n", (int) A[j].size, p);
      p+= c;
      c= 1;
      j= i;
    }
  }
  if (j==nA-1) fprintf(fA, "%d %d\n", (int) A[j].size, p+c);
  fclose(fA);
}


int main(int argc,char **argv)
{
  bff_t *bf;
  int i,N,op,do_check=0,do_dump=0;
  FILE *plo;

  bf= bff_new();
  seed_random(&argc,argv);
  N= atoi(argv[1]);
  plo= NULL;

  for(i=2;i<argc;i++) if (!strcmp(argv[i],"-c")) { do_check=1; break; }
  for(i=2;i<argc;i++) if (!strcmp(argv[i],"-v")) { do_dump=1; break; }
  for(i=2;i<argc;i++) if (!strcmp(argv[i],"-p")) { plo= fopen("data.txt","w"); break; }
  for(i=0;i<N;i++)
  {
    double pp;
    pp= (double) rand() / RAND_MAX;
    // op= rand() % 100;
    if (pp<0.3) do_insert(bf);
    else if (pp<0.6) do_remove(bf);
    else do_find(bf);

    if (do_check) check_file(bf);
    if (do_dump) { printf("DUMP----\n"); dump_objects(bf); dump_free(bf); }
    if (plo)
     fprintf(plo, "%d %d %d %d\n", i, 
              (int) ((double)obsize*100/bf->file_size),
                    nTob, (int) bf->file_size/1024);
  }
  printf("useful data size= %d\n", obsize);
  printf("file size= %d\n", (int) bf->file_size);
  printf("utilization is %d%%\n", (int) ((double)obsize*100/bf->file_size));
  printf("max # objects was: %d\n", maxnob);
  printf("%d insertions %d deletions %d lookups\n", Ni, Nd, Nf);

  for(i=2;i<argc;i++) if (!strcmp(argv[i],"-d")) { dump_objects(bf); break; }
  for(i=2;i<argc;i++) if (!strcmp(argv[i],"-f")) { dump_free(bf); break; }
  for(i=2;i<argc;i++) if (!strcmp(argv[i],"-a")) { analyse(bf); break; }

  if (plo) fclose(plo);

  return 0;
}

